# Copyright 2016-2020, Pulumi Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import math
import os
import re
import subprocess
from random import random
from typing import Optional, Union, List, Mapping, Sequence
from distutils.version import LooseVersion

import pulumi
from .utils import get_image_name_and_tag


class Registry:
    registry: pulumi.Input[str]
    """
    Registry server url
    """
    username: pulumi.Input[str]
    """
    Username for the registry
    """
    password: pulumi.Input[str]
    """
    Password for the registry
    """

    def __init__(self, registry: pulumi.Input[str], username: pulumi.Input[str], password: pulumi.Input[str]):
        """
        Registry is the information required to login to a Docker registry.


        :param pulumi.Input[str] registry: Registry server url
        :param pulumi.Input[str] username: Username for the registry
        :param pulumi.Input[str] password: Password for the registry
        """
        self.registry = registry
        self.username = username
        self.password = password


class CacheFrom:
    """
    CacheFrom may be used to specify build stages to use for the Docker build cache. The final image
    is always implicitly included.
    """

    stages: Optional[Sequence[pulumi.Input[pulumi.Input[str]]]]
    """
    An optional list of build stages to use for caching. Each build stage in this list will be
    built explicitly and pushed to the target repository. A given stage's image will be tagged as
    "[stage-name]".
    """


class DockerBuild:
    context: Optional[pulumi.Input[str]]
    """
    context is a path to a directory to use for the Docker build context, usually the directory
    in which the Dockerfile resides (although dockerfile may be used to choose a custom location
    independent of this choice). If not specified, the context defaults to the current working
    directory if a relative path is used, it is relative to the current working directory that
    Pulumi is evaluating.
    """

    dockerfile: Optional[pulumi.Input[str]]
    """
    dockerfile may be used to override the default Dockerfile name and/or location.  By default,
    it is assumed to be a file named Dockerfile in the root of the build context.
    """

    args: Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]]
    """
    An optional map of named build-time argument variables to set during the Docker build.  This
    flag allows you to pass built-time variables that can be accessed like environment variables
    inside the `RUN` instruction.
    """

    cache_from: Optional[pulumi.Input[Union[bool, CacheFrom]]]
    """
    An optional CacheFrom object with information about the build stages to use for the Docker
    build cache. This parameter maps to the --cache-from argument to the Docker CLI. If this
    parameter is `true`, only the final image will be pulled and passed to --cache-from if it is
    a CacheFrom object, the stages named therein will also be pulled and passed to --cache-from.
    """

    extra_options: Optional[Sequence[pulumi.Input[pulumi.Input[str]]]]
    """
    An optional catch-all str to provide extra CLI options to the docker build command.  For
    example, use to specify `--network host`.
    """

    env: Optional[Mapping[str, str]]
    """
    Environment variables to set on the invocation of `docker build`, for example to support
    `DOCKER_BUILDKIT=1 docker build`.
    """

    target: Optional[pulumi.Input[str]]
    """
    The target of the dockerfile to build
    """

    def __init__(self, context=None, dockerfile=None, args=None, cache_from=None, extra_options=None, env=None,
                 target=None):
        """
        DockerBuild may be used to specify detailed instructions about how to build a container.


        :param Optional[pulumi.Input[str]] context: context is a path to a directory to use for the Docker build
            context, usually the directory in which the Dockerfile resides (although dockerfile may be used to choose
            a custom location independent of this choice). If not specified, the context defaults to the current working
            directory if a relative path is used, it is relative to the current working directory that
            Pulumi is evaluating.
        :param Optional[pulumi.Input[str]] dockerfile: dockerfile may be used to override the default Dockerfile name
            and/or location.  By default, it is assumed to be a file named Dockerfile in the root of the build context.
        :param Optional[pulumi.Input[Mapping[str, pulumi.Input[str]]]] args: An optional map of named build-time
            argument variables to set during the Docker build.  This flag allows you to pass built-time variables that
            can be accessed like environment variables inside the `RUN` instruction.
        :param Optional[pulumi.Input[Union[bool, CacheFrom]]] cache_from: An optional CacheFrom object with information
            about the build stages to use for the Docker build cache. This parameter maps to the --cache-from argument
            to the Docker CLI. If this parameter is `true`, only the final image will be pulled and passed to
            --cache-from if it is a CacheFrom object, the stages named therein will also be pulled and passed to
            --cache-from.
        :param Optional[Sequence[pulumi.Input[pulumi.Input[str]]]] extra_options: An optional catch-all str to provide
            extra CLI options to the docker build command.  For example, use to specify `--network host`.
        :param Optional[Mapping[str, str]] env: Environment variables to set on the invocation of `docker build`, for
         example to support `DOCKER_BUILDKIT=1 docker build`.
        :param Optional[pulumi.Input[str]] target: The target of the dockerfile to build
        """
        self.context = context
        self.dockerfile = dockerfile
        self.args = args
        self.cache_from = cache_from
        self.extra_options = extra_options
        self.env = env
        self.target = target


class Error(Exception):
    pass


class ResourceError(Error):
    def __init__(self, message: str, resource: Optional[pulumi.Resource], hide_stack: Optional[bool] = False):
        self.resource = resource
        self.hide_stack = hide_stack
        super().__init__(message)


def use_docker_password_stdin(log_resource: pulumi.Resource):
    # Verify that 'docker' is on the PATH and get the client/server versions
    try:
        docker_version_str = run_command_that_must_succeed(
            "docker", ["version", "-f", "{{json .}}"], log_resource)
        # IDEA: In the future we could warn here on out-of-date versions of Docker which may not support key
        # features we want to use.

        pulumi.log.debug(f'\'docker version\' => {docker_version_str}', log_resource)
    except Exception:
        raise ResourceError("No 'docker' command available on PATH: Please install to use container 'build' mode.",
                            log_resource)

    # Decide whether to use --password or --password-stdin based on the client version.
    try:
        version_data: any = json.loads(docker_version_str)
        client_version: str = version_data['Client']['Version']
        return LooseVersion(client_version) >= LooseVersion("17.07.0")
    except Exception as err:
        pulumi.log.info(f'Could not process Docker version ({err})', log_resource)
    return False


def build_and_push_image(
    base_image_name: str,
    path_or_build: pulumi.Input[Union[str, DockerBuild]],
    repository_url: pulumi.Input[str],
    log_resource: pulumi.Resource,
    registry: Optional[Registry],
    skip_push: bool = False
) -> str:
    """
    build_and_push_image will build and push the Dockerfile and context from [pathOrBuild] into the
    requested docker repo [repository_url].  It returns the unique target image name for the image in
    the docker repository.  During preview this will build the image, and return the target image
    name, without pushing. During a normal update, it will do the same, as well as tag and push the
    image.
    """

    # Give an initial message indicating what we're about to do.  That way, if anything
    # takes a while, the user has an idea about what's going on.
    log_ephemeral("Starting docker build and push...", log_resource)

    check_repository_url(repository_url)

    _, tag = get_image_name_and_tag(base_image_name)

    # login immediately if we're going to have to actually communicate with a remote registry.
    #
    # We know we have to login if:
    #
    #  1. We're doing an update.  In that case, we'll always want to login so we can push our
    #     images to the remote registry.
    #
    # 2. We're in preview or update and the build information contains 'cache from' information. In
    #    that case, we'll want want to pull from the registry and will need to login for that.

    pull_from_cache = not isinstance(path_or_build,
                                     str) and path_or_build and path_or_build.cache_from and repository_url is None

    # If no `registry` info was passed in we simply assume docker is already
    # logged-in to the correct registry (or uses auto-login via credential helpers).
    if registry:
        if not pulumi.runtime.is_dry_run() or pull_from_cache:
            log_ephemeral("Logging in to registry...", log_resource)
            login_to_registry(registry, log_resource)

    # If the container specified a cache_from parameter, first set up the cached stages.
    cache_from = None
    if pull_from_cache:
        _cache_from_param = CacheFrom() if isinstance(path_or_build.cache_from, bool) else path_or_build.cache_from
        cache_from_param = _cache_from_param if _cache_from_param else CacheFrom()
        cache_from = pull_cache(base_image_name, cache_from_param, repository_url, log_resource)

    # Next, build the image.
    build_result = build_image(base_image_name, path_or_build, log_resource, cache_from)
    image_id, stages = build_result.image_id, build_result.stages

    if image_id is None:
        raise Error("Internal error: docker build did not produce an imageId.")

    # Generate a name that uniquely will identify this built image.  This is similar in purpose to
    # the name@digest form that can be normally be retrieved from a docker repository.  However,
    # this tag doesn't require actually pushing the image, nor does it require communicating with
    # some external system, making it suitable for unique identification, even during preview.
    # This also means that if docker produces a new imageId, we'll get a new name here, ensuring that
    # resources (like docker.Image and cloud.Service) will be appropriately replaced.
    unique_tagged_image_name = create_tagged_image_name(repository_url, tag, image_id)

    # Use those to push the image.  Then just return the unique target name. as the final result
    # for our caller to use. Only push the image during an update, do not push during a preview.
    if not pulumi.runtime.is_dry_run() and not skip_push:
        # Push the final image first, then push the stage images to use for caching.

        # First, push with both the optionally-requested-tag *and* imageId (which is guaranteed to
        # be defined).  By using the imageId we give the image a fully unique location that we can
        # successfully pull regardless of whatever else has happened at this repository_url.

        # Next, push only with the optionally-requested-tag.  Users of this API still want to get a
        # nice and simple url that they can reach this image at, without having the explicit imageId
        # hash added to it.  Note: this location is not guaranteed to be idempotent.  For example,
        # pushes on other machines might overwrite that location.
        tag_and_push_image(base_image_name, repository_url, tag, image_id, log_resource=log_resource)
        tag_and_push_image(base_image_name, repository_url, tag, image_id=None, log_resource=log_resource)

        for stage in stages:
            tag_and_push_image(
                local_stage_image_name(base_image_name, stage),
                repository_url,
                stage,
                image_id=None,
                log_resource=log_resource
            )

    # If we got here, then building/pushing didn't throw any errors.  Update the status bar
    # indicating that things worked properly.  That way, the info bar isn't stuck showing the very
    # last thing printed by some subcommand we launched.
    log_ephemeral("Successfully pushed to docker", log_resource)

    return unique_tagged_image_name


def log_ephemeral(message: str, log_resource: pulumi.Resource):
    try:
        pulumi.log.info(message, log_resource, stream_id=None, ephemeral=True)
    except TypeError:
        # that pulumi version does not support ephemeral
        pulumi.log.info(message, log_resource, stream_id=None)


def check_repository_url(repository_url: str):
    _, tag = get_image_name_and_tag(repository_url)

    # We want to report an advisory error to users so that they don't accidentally include a 'tag'
    # in the repo url they supply.  i.e. their repo url can be:
    #
    #      docker.mycompany.com/namespace/myimage
    #
    # but should not be:
    #
    #      docker.mycompany.com/namespace/myimage:latest
    #
    # We could consider removing this check entirely.  However, it is likely valuable to catch
    # clear mistakes where a tag was included in a repo url inappropriately.
    #
    # However, since we do have the check, we need to ensure that we do allow the user to specify
    # a *port* on their repository that the are communicating with.  i.e. it's fine to have:
    #
    #      docker.mycompany.com:5000 or
    #      docker.mycompany.com:5000/namespace/myimage
    #
    # So check if this actually does look like a port, and don't report an error in that case.
    #
    # From: https:#www.w3.org/Addressing/URL/url-spec.txt
    #
    #      port        digits
    #
    # Regex = any number of digits, optionally followed by / and any remainder.

    if tag and not re.match(r'^\d+(/.*)?', tag):
        raise Error(f'[repository_url] should not contain a tag: {tag}')


def local_stage_image_name(image_name: str, stage: str):
    return f'{image_name}-{stage}'


def create_tagged_image_name(repository_url: str, tag: Optional[str], image_id: Optional[str]) -> str:
    pieces: List = []
    if tag:
        pieces.append(tag)

    if image_id:
        pieces.append(image_id)

    # Note: we don't do any validation that the tag is well formed, as per:
    # https://docs.docker.com/engine/reference/commandline/tag
    #
    # If there are any issues with it, we'll just let docker report the problem.
    full_tag = "-".join(pieces)
    return f'{repository_url}:{full_tag}' if full_tag else repository_url


def pull_cache(
    image_name: str,
    cache_from,
    repo_url: str,
    log_resource: pulumi.Resource
) -> Optional[Sequence[str]]:
    # Ensure that we have a repository URL. If we don't, we won't be able to pull anything.
    if not repo_url:
        return None

    pulumi.log.debug(f'pulling cache for {image_name} from {repo_url}', log_resource)

    cache_from_images: List = []
    stages = (cache_from.stages if cache_from.stages else []).concat([""])
    for stage in stages:
        tag = f':{stage}' if stage else ""
        image = f'{repo_url}{tag}'

        # Try to pull the existing image if it exists.  This may fail if the image does not exist.
        # That's fine, just move onto the next stage.  Also, pass along a flag saying that we
        # should print that error as a warning instead.  We don't want the update to succeed but
        # the user to then get a nasty "error:" message at the end.
        code, _ = run_command_that_can_fail(
            "docker", ["pull", image], log_resource,
            report_full_command_line=True, report_error_as_warning=True
        )
        if code:
            continue

        cache_from_images.append(image)

    return cache_from_images


class BuildResult:
    image_id: str
    stages: Sequence[str]

    def __init__(self, image_id, stages):
        self.image_id = image_id
        self.stages = stages


def build_image(
    image_name: str,
    path_or_build: Union[str, DockerBuild],
    log_resource: pulumi.Resource,
    cache_from: Optional[str]
) -> BuildResult:
    if isinstance(path_or_build, str):
        build = DockerBuild(context=path_or_build)
    elif path_or_build:
        build = path_or_build
    else:
        raise ResourceError(f'Cannot build a container with an empty build specification', log_resource)

    # If the build context is missing, default it to the working directory.
    if not build.context:
        build.context = "."

    log_ephemeral(
        f'Building container image \'{image_name}\': context={build.context}' +
        (f', dockerfile={build.dockerfile}' if build.dockerfile else "") +
        (f', args={json.dumps(build.args)}' if build.args else "") +
        (f', target={build.target}' if build.target else ""), log_resource)

    # If the container build specified build stages to cache, build each in turn.
    stages = []
    if build.cache_from and not isinstance(build.cache_from, bool) and build.cache_from.stages:
        for stage in build.cache_from.stages:
            docker_build(
                local_stage_image_name(image_name, stage), build,
                cache_from=cache_from, log_resource=log_resource, target=stage)
            stages.append(stage)

    # Invoke Docker CLI commands to build.
    docker_build(image_name, build, log_resource, cache_from)

    # Finally, inspect the image so we can return the SHA digest. Do not forward the output of this
    # command this to the CLI to show the user.
    inspect_result = run_command_that_must_succeed(
        "docker", ["image", "inspect", "-f", "{{.Id}}", image_name], log_resource)
    if not inspect_result:
        raise ResourceError(
            f'No digest available for image {image_name}', log_resource)

    # From https:#docs.docker.com/registry/spec/api/#content-digests
    #
    # the image id will be a "algorithm:hex" pair.  We don't care about the algorithm part.  All we
    # want is the unique portion we can use elsewhere.  Since we are also going to place this in an
    # image tag, we also don't want the colon, as that's not legal there.  So simply grab the hex
    # portion after the colon and return that.

    image_id = inspect_result.strip()
    colon_index = image_id.rfind(":")
    image_id = image_id if colon_index < 0 else image_id[colon_index + 1:]

    return BuildResult(image_id, stages)


def docker_build(
    image_name: str,
    build: DockerBuild,
    log_resource: pulumi.Resource,
    cache_from: Optional[str],
    target: Optional[str] = None
) -> str:
    # Prepare the build arguments.
    build_args: List[str] = ["build"]
    if build.dockerfile:
        build_args.extend(["-f", build.dockerfile])  # add a custom Dockerfile location.

    if build.args:
        for arg, build_arg in build.args.items():
            build_args.extend(["--build-arg", f'{arg}={build_arg}'])
    if build.target:
        build_args.extend(["--target", build.target])
    if build.cache_from:
        cache_from_images = cache_from
        if cache_from_images and cache_from_images.length:
            build_args.extend(["--cache-from", ''.join(cache_from_images)])
    if build.extra_options:
        build_args.extend(build.extra_options)

    build_args.extend(["-t", image_name])  # tag the image with the chosen name.
    if target:
        build_args.extend(["--target", target])

    if build.context:
        build_args.append(build.context)  # push the docker build context onto the path.

    return run_command_that_must_succeed("docker", build_args, log_resource, env=build.env)


class LoginResult:
    registry_name: str
    username: str

    def __init__(self, registry_name: str, username: str):
        self.registry_name = registry_name
        self.username = username


# Keep track of registries and users that have been logged in.  If we've already logged into that
# registry with that user, there's no need to do it again.
login_results: List[LoginResult] = []


def login_to_registry(registry: Registry, log_resource: pulumi.Resource):
    registry_name = registry.registry
    username = registry.username
    password = registry.password

    # See if we've issued an outstanding requests to login into this registry.  If so, just
    # await the results of that login request.  Otherwise, create a new request and keep it
    # around so that future login requests will see it.
    for result in login_results:
        if result.registry_name == registry_name and result.username == username:
            log_ephemeral(f'Reusing existing login for {username}@{registry_name}', log_resource)
            return

    docker_password_stdin = use_docker_password_stdin(log_resource)

    # pass 'report_full_command_line: false' here so that if we fail to login we don't emit the
    # username/password in our logs.  Instead, we'll just say "'docker login' failed with code ..."
    if docker_password_stdin:
        run_command_that_must_succeed("docker", ["login", "-u", username, "--password-stdin", registry_name],
                                      log_resource, report_full_command_line=False, stdin=password)
    else:
        run_command_that_must_succeed("docker", ["login", "-u", username, "-p", password, registry_name],
                                      log_resource, report_full_command_line=False)

    login_results.append(LoginResult(registry_name, username))


def tag_and_push_image(
    image_name: str, repository_url: str,
    tag: Optional[str], image_id: Optional[str],
    log_resource: pulumi.Resource
):
    def do_tag_and_push(target_name: str):
        run_command_that_must_succeed("docker", ["tag", image_name, target_name], log_resource)
        run_command_that_must_succeed("docker", ["push", target_name], log_resource)

    # Ensure we have a unique target name for this image, and tag and push to that unique target.
    do_tag_and_push(create_tagged_image_name(repository_url, tag, image_id))

    # If the user provided a tag themselves (like "x/y:dev") then also tag and push directly to
    # that 'dev' tag.  This is not going to be a unique location, and future pushes will overwrite
    # this location.  However, that's ok as there's still the unique target we generated above.
    #
    # Note: don't need to do this if imageId was 'undefined' as the above line will have already
    # taken care of things for us.
    if tag is not None and image_id is not None:
        do_tag_and_push(create_tagged_image_name(repository_url, tag, image_id=None))

    return


class CommandResult:
    code: int
    stdout: str

    def __init__(self, code, stdout):
        self.code = code
        self.stdout = stdout


def get_command_line_message(
    cmd: str, args: Sequence[str], report_full_command_line: bool, env: Optional[Mapping[str, str]] = None
):
    elements = []
    if env:
        elements.append(" ".join(map(lambda k: f'{k}={env[k]}', env.keys())))
    elements.append(cmd)

    argstr = " ".join(args) if report_full_command_line else args[0]
    elements.append(argstr)
    return f"'{' '.join(elements)}'"


def get_failure_message(
    cmd: str, args: Sequence[str], report_full_command_line: bool, code: int, env: Optional[Mapping[str, str]] = None
):
    return f'{get_command_line_message(cmd, args, report_full_command_line, env)} failed with exit code {code}'


# [report_full_command_line] is used to determine if the full command line should be reported
# when an error happens.  In general reporting the full command line is fine.  But it should be set
# to false if it might contain sensitive information (like a username/password)
def run_command_that_must_succeed(
    cmd: str,
    args: Sequence[str],
    log_resource: pulumi.Resource,
    report_full_command_line: bool = True,
    stdin: Optional[str] = None,
    env: Optional[Mapping] = None
) -> str:
    command_result = run_command_that_can_fail(
        cmd, args, log_resource, report_full_command_line, False, stdin, env)
    code, stdout = command_result.code, command_result.stdout

    if code != 0:
        # Fail the entire build and push.  This includes the full output of the command so that at
        # the end the user can review the full docker message about what the problem was.
        #
        # Note: a message about the command failing will have already been ephemerally reported to
        # the status column.
        raise ResourceError(
            f'{get_failure_message(cmd, args, report_full_command_line, code)}\n{stdout}', log_resource)

    return stdout


def run_command_that_can_fail(
    cmd_name: str,
    args: Sequence[str],
    log_resource: pulumi.Resource,
    report_full_command_line: bool,
    report_error_as_warning: bool,
    stdin: Optional[str] = None,
    env: Optional[Mapping[str, str]] = None
) -> CommandResult:
    """
    Runs a CLI command in a child process, returning a future for the process's exit. Both stdout
    and stderr are redirected to process.stdout and process.stderr by default.

    If the [stdin] argument is defined, its contents are piped into stdin for the child process.

    [log_resource] is used to specify the resource to associate command output with. Stderr messages
    are always sent (since they may contain important information about something that's gone wrong).
    Stdout messages will be logged ephemerally to this resource.  This lets the user know there is
    progress, without having that dumped on them at the end.  If an error occurs though, the stdout
    content will be printed.
    """

    # Let the user ephemerally know the command we're going to execute.
    log_ephemeral(f"Executing {get_command_line_message(cmd_name, args, report_full_command_line, env)}", log_resource)

    # Generate a unique stream-ID that we'll associate all the docker output with. This will allow
    # each spawned CLI command's output to associated with 'resource' and also streamed to the UI
    # in pieces so that it can be displayed live.  The stream-ID is so that the UI knows these
    # messages are all related and should be considered as one large message (just one that was
    # sent over in chunks).
    #
    # We use Math.random here in case our package is loaded multiple times in memory (i.e. because
    # different downstream dependencies depend on different versions of us).  By being random we
    # effectively make it completely unlikely that any two cli outputs could map to the same stream
    # id.
    #
    # Pick a reasonably distributed number between 0 and 2^30.  This will fit as an int32
    # which the grpc layer needs.
    stream_id = math.floor(random() * (1 << 30))

    cmd = [cmd_name]
    cmd.extend(args)

    if env is not None:
        env = os.environ.copy().update(env)

    process = subprocess.Popen(cmd, env=env, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, stdin=subprocess.PIPE, encoding="utf-8")

    # We store the results from stdout in memory and will return them as a str.
    stdout_chunks: List[str] = []
    stderr_chunks: List[str] = []

    # A None value for process.returncode indicates that the process hasn't terminated yet
    while process.returncode is None:
        outs, errs = process.communicate(input=stdin)
        if outs:
            # Report all stdout messages as ephemeral messages.  That way they show up in the
            # info bar as they're happening.  But they do not overwhelm the user as the end
            # of the run.
            log_ephemeral(outs, log_resource)
            stdout_chunks.append(outs.rstrip())
        if errs:
            # We can't stream these stderr messages as we receive them because we don't knows at
            # this point because Docker uses stderr for both errors and warnings.  So, instead, we
            # just collect the messages, and wait for the process to end to decide how to report
            # them.
            stderr_chunks.append(errs.rstrip())

    code = process.returncode

    # Collapse our stored stdout/stderr messages into single strings.
    stderr = ''.join(stderr_chunks)
    stdout = ''.join(stdout_chunks)

    # If we got any stderr messages, report them as an error/warning depending on the
    # result of the operation.
    if stderr:
        if code and not report_error_as_warning:
            # Command returned non-zero code.  Treat these stderr messages as an error.
            pulumi.log.error(stderr, log_resource, stream_id)
        else:
            # command succeeded.  These were just warning.
            pulumi.log.warn(stderr, log_resource, stream_id)

    # If the command failed report an ephemeral message indicating which command it was.
    # That way the user can immediately see something went wrong in the info bar.  The
    # caller (normally run_command_that_can_succeed) can choose to also report this
    # non-ephemerally.
    if code:
        log_ephemeral(get_failure_message(cmd_name, args, report_full_command_line, code), log_resource)

    return CommandResult(code, stdout)



# Copyright 2016-2020, Pulumi Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pulumi
from typing import Optional, Union

from .docker import build_and_push_image, DockerBuild, Registry
from .utils import get_image_name_and_tag


class ImageRegistry:
    server: pulumi.Input[str]
    """
    Docker registry server URL to push to. 
    
    Some common values include:
        - DockerHub: `docker.io` or `https://index.docker.io/v1`
        - Azure Container Registry: `<name>.azurecr.io`
        - AWS Elastic Container Registry: `<account>.dkr.ecr.us-east-2.amazonaws.com`
        - Google Container Registry: `<name>.gcr.io`
    """

    username: pulumi.Input[str]
    """
    Username for login to the target Docker registry.
    """

    password: pulumi.Input[str]
    """
    Password for login to the target Docker registry.
    """

    def __init__(self, server: pulumi.Input[str], username: pulumi.Input[str], password: pulumi.Input[str]):
        """
        Image is a resource that represents a Docker image built locally which is published and made
        available via a remote Docker registry.  This can be used to ensure that a Docker source
        directory from a local deployment environment is built and pushed to a cloud-hosted Docker
        registry as part of a Pulumi deployment, so that it can be referenced as an image input from
        other cloud services that reference Docker images - including Kubernetes Pods, AWS ECS Tasks, and
        Azure Container Instances.


        :param pulumi.Input[str] server: Docker registry server URL to push to.
        :param pulumi.Input[str] username: Username for login to the target Docker registry.
        :param pulumi.Input[str] password: Password for login to the target Docker registry.
        """
        self.server = server
        self.username = username
        self.password = password


class _ImageArgs:
    image_name: str
    """
    The qualified image name that will be pushed to the remote registry.  Must be a supported
    image name for the target registry user.  This name can include a tag at the end.  If
    provided all pushed image resources will contain that tag as well.

    Either [image_name] or [localImageName] can have a tag.  However, if both have a tag, then
    those tags must match.
    """

    build: Union[str, DockerBuild]
    """
    The Docker build context, as a folder path or a detailed DockerBuild object.
    """

    local_image_name: str
    """
    The docker image name to build locally before tagging with image_name.  If not provided, it
    will be given the value of to [image_name].  This name can include a tag at the end.  If
    provided all pushed image resources will contain that tag as well.

    Either [image_name] or [localImageName] can have a tag.  However, if both have a tag, then
    those tags must match.
    """

    registry: Optional[ImageRegistry]
    """
    Credentials for the docker registry to push to.
    """

    skip_push: Optional[bool]
    """
    Skip push flag.
    """

    def __init__(self,
                 image_name: str,
                 build: Union[str, DockerBuild],
                 local_image_name: str,
                 registry_server: Optional[str],
                 registry_username: Optional[str],
                 registry_password: Optional[str],
                 skip_push: Optional[bool]):
        """
        Arguments for constructing an Image resource.


        :param str image_name: The qualified image name that will be pushed to the remote registry.  Must be a supported
            image name for the target registry user.  This name can include a tag at the end.  If
            provided all pushed image resources will contain that tag as well.
            Either [image_name] or [localImageName] can have a tag.  However, if both have a tag, then
            those tags must match.
        :param Union[str, DockerBuild] build: The Docker build context, as a folder path or a detailed DockerBuild
            object.
        :param str local_image_name: The docker image name to build locally before tagging with image_name.  If not
            provided, it will be given the value of to [image_name].  This name can include a tag at the end.  If
            provided all pushed image resources will contain that tag as well.
            Either [image_name] or [localImageName] can have a tag.  However, if both have a tag, then
            those tags must match.
        :param Optional[str] registry_server: Docker registry to push to.
        :param Optional[str] registry_username:  Username for the registry.
        :param Optional[str] registry_password: Password for the registry.
        :param Optional[bool]) skip_push: Skip push flag.
        """
        self.image_name = image_name
        self.build = build
        self.local_image_name = local_image_name
        self.registry_server = registry_server
        self.registry_username = registry_username
        self.registry_password = registry_password
        self.skip_push = skip_push


class Image(pulumi.ComponentResource):
    base_image_name: pulumi.Output[str]
    """
    The base image name that was built and pushed.  This does not include the id annotation, so
    is not pinned to the specific build performed by this docker.Image.
    """

    image_name: pulumi.Output[str]
    """
    The unique pinned image name on the remote repository."""

    registry_server: pulumi.Output[Optional[str]]
    """
    The server the image is located at.
    """

    def __init__(self, name: str,
                 image_name: pulumi.Input[str],
                 build: Union[pulumi.Input[str], DockerBuild],
                 local_image_name: Optional[pulumi.Input[str]] = None,
                 registry: Optional[pulumi.Input[ImageRegistry]] = None,
                 skip_push: Optional[bool] = None,
                 opts: Optional[pulumi.ResourceOptions] = None):
        """
        A docker.Image resource represents a Docker image built locally which is published and made
        available via a remote Docker registry.  This can be used to ensure that a Docker source
        directory from a local deployment environment is built and pushed to a cloud-hosted Docker
        registry as part of a Pulumi deployment, so that it can be referenced as an image input from
        other cloud services that reference Docker images - including Kubernetes Pods, AWS ECS Tasks, and
        Azure Container Instances.


        :param str name: The name of the resource.
        :param pulumi.Input[str] image_name: The qualified image name that will be pushed to the remote registry.
            Must be a supported image name for the target registry user.  This name can include a tag at the end.  If
            provided all pushed image resources will contain that tag as well.
            Either [imageName] or [localImageName] can have a tag.  However, if both have a tag, then
            those tags must match.
        :param Union[pulumi.Input[str], DockerBuild] build: The Docker build context, as a folder path or a detailed
            DockerBuild object.
        :param Optional[pulumi.Input[str]] local_image_name: The docker image name to build locally before tagging with
            imageName.  If not provided, it will be given the value of to [imageName].  This name can include a tag at
            the end.  If provided all pushed image resources will contain that tag as well.
            Either [imageName] or [localImageName] can have a tag.  However, if both have a tag, then
            those tags must match.
        :param Optional[pulumi.Input[ImageRegistry]] registry: Credentials for the docker registry to push to.
        :param Optional[bool] skip_push: Skip push flag.
        :param Optional[pulumi.ResourceOptions] opts: Options for the resource.
        """
        super().__init__("docker:image:Image", name, {}, opts)

        def get_image_data(image_args: _ImageArgs):
            image_name = image_args.image_name

            # If there is no local_image_name set it equal to image_name.  Note: this means
            # that if image_name contains a tag, local_image_name will contain the same tag.
            local_image_name = image_args.local_image_name or image_name

            # Now break both the localImageName and the image_name into the untagged part and the
            # optional tag.  If both have tags, they must match.  If one or the other has a tag, we
            # just use that as the tag to use.  This allows users to flexibly provide a tag on one
            # option or the other and still have it work out.
            local_image_name_without_tag, local_image_name_tag = get_image_name_and_tag(local_image_name)
            image_name_without_tag, image_name_tag = get_image_name_and_tag(image_name)

            tag = local_image_name_tag or image_name_tag

            def check_tag(t: Optional[str]):
                if t and (t != tag):
                    raise Exception(f'[local_image_name] and [image_name]'
                                    ' had mismatched tags. {local_image_name_tag} != {image_name_tag}')

            check_tag(local_image_name_tag)
            check_tag(image_name_tag)

            # build_and_push_image expects only the base_image_name to have a tag.  So build that
            # name appropriately if we were given a tag.
            base_image_name = f'{local_image_name_without_tag}:{tag}' if tag else local_image_name

            # build_and_push_image does not want the repository_url to have a tag.  This is just
            # the base url where the images will be pushed to.  All tagging will be taken care of
            # inside that api.
            repository_url = image_name_without_tag

            def get_registry():
                registry_server = image_args.registry_server
                registry_password = image_args.registry_password
                registry_username = image_args.registry_username

                if registry_server and registry_username and registry_password:
                    return Registry(registry_server, registry_username, registry_password)
                else:
                    return None

            unique_target_name = build_and_push_image(
                base_image_name,
                image_args.build,
                repository_url,
                self,
                get_registry(),
                image_args.skip_push,
            )

            return {
                'image_name': unique_target_name,
                'registry_server': image_args.registry_server or None
            }

        image_data = pulumi.Output.all(
            image_name,
            build,
            local_image_name,
            registry.server if registry else None,
            registry.username if registry else None,
            registry.password if registry else None,
            skip_push,
        ).apply(
            lambda args: get_image_data(_ImageArgs(*args))
        )

        self.image_name, self.registry_server = image_data['image_name'], image_data['registry_server']

        self.base_image_name = pulumi.Output.from_input(image_name)

        self.register_outputs({
            'base_image_name': self.image_name,
            'image_name': self.image_name,
            'registry_server': self.registry_server,
        })



# coding=utf-8
# *** WARNING: this file was generated by the Pulumi Terraform Bridge (tfgen) Tool. ***
# *** Do not edit by hand unless you're certain you know what you are doing! ***

import warnings
import pulumi
import pulumi.runtime
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
from . import _utilities, _tables
from . import outputs

__all__ = [
    'GetNetworkResult',
    'AwaitableGetNetworkResult',
    'get_network',
]

@pulumi.output_type
class GetNetworkResult:
    """
    A collection of values returned by getNetwork.
    """
    def __init__(__self__, driver=None, id=None, internal=None, ipam_configs=None, name=None, options=None, scope=None):
        if driver and not isinstance(driver, str):
            raise TypeError("Expected argument 'driver' to be a str")
        pulumi.set(__self__, "driver", driver)
        if id and not isinstance(id, str):
            raise TypeError("Expected argument 'id' to be a str")
        pulumi.set(__self__, "id", id)
        if internal and not isinstance(internal, bool):
            raise TypeError("Expected argument 'internal' to be a bool")
        pulumi.set(__self__, "internal", internal)
        if ipam_configs and not isinstance(ipam_configs, list):
            raise TypeError("Expected argument 'ipam_configs' to be a list")
        pulumi.set(__self__, "ipam_configs", ipam_configs)
        if name and not isinstance(name, str):
            raise TypeError("Expected argument 'name' to be a str")
        pulumi.set(__self__, "name", name)
        if options and not isinstance(options, dict):
            raise TypeError("Expected argument 'options' to be a dict")
        pulumi.set(__self__, "options", options)
        if scope and not isinstance(scope, str):
            raise TypeError("Expected argument 'scope' to be a str")
        pulumi.set(__self__, "scope", scope)

    @property
    @pulumi.getter
    def driver(self) -> str:
        """
        (Optional, string) The driver of the Docker network. 
        Possible values are `bridge`, `host`, `overlay`, `macvlan`.
        See [docker docs][networkdocs] for more details.
        """
        return pulumi.get(self, "driver")

    @property
    @pulumi.getter
    def id(self) -> Optional[str]:
        return pulumi.get(self, "id")

    @property
    @pulumi.getter
    def internal(self) -> bool:
        return pulumi.get(self, "internal")

    @property
    @pulumi.getter(name="ipamConfigs")
    def ipam_configs(self) -> List['outputs.GetNetworkIpamConfigResult']:
        return pulumi.get(self, "ipam_configs")

    @property
    @pulumi.getter
    def name(self) -> Optional[str]:
        return pulumi.get(self, "name")

    @property
    @pulumi.getter
    def options(self) -> Mapping[str, Any]:
        """
        (Optional, map) Only available with bridge networks. See
        [docker docs][bridgeoptionsdocs] for more details.
        * `internal` (Optional, bool) Boolean flag for whether the network is internal.
        * `ipam_config` (Optional, map) See IPAM below for details.
        * `scope` (Optional, string) Scope of the network. One of `swarm`, `global`, or `local`.
        """
        return pulumi.get(self, "options")

    @property
    @pulumi.getter
    def scope(self) -> str:
        return pulumi.get(self, "scope")


class AwaitableGetNetworkResult(GetNetworkResult):
    # pylint: disable=using-constant-test
    def __await__(self):
        if False:
            yield self
        return GetNetworkResult(
            driver=self.driver,
            id=self.id,
            internal=self.internal,
            ipam_configs=self.ipam_configs,
            name=self.name,
            options=self.options,
            scope=self.scope)


def get_network(id: Optional[str] = None,
                name: Optional[str] = None,
                opts: Optional[pulumi.InvokeOptions] = None) -> AwaitableGetNetworkResult:
    """
    Finds a specific docker network and returns information about it.

    ## Example Usage

    ```python
    import pulumi
    import pulumi_docker as docker

    main = docker.get_network(name="main")
    ```


    :param str id: The id of the Docker network.
    :param str name: The name of the Docker network.
    """
    __args__ = dict()
    __args__['id'] = id
    __args__['name'] = name
    if opts is None:
        opts = pulumi.InvokeOptions()
    if opts.version is None:
        opts.version = _utilities.get_version()
    __ret__ = pulumi.runtime.invoke('docker:index/getNetwork:getNetwork', __args__, opts=opts, typ=GetNetworkResult).value

    return AwaitableGetNetworkResult(
        driver=__ret__.driver,
        id=__ret__.id,
        internal=__ret__.internal,
        ipam_configs=__ret__.ipam_configs,
        name=__ret__.name,
        options=__ret__.options,
        scope=__ret__.scope)



# coding=utf-8
# *** WARNING: this file was generated by the Pulumi Terraform Bridge (tfgen) Tool. ***
# *** Do not edit by hand unless you're certain you know what you are doing! ***

# Export this package's modules as members:
from .container import *
from .docker import *
from .get_network import *
from .get_registry_image import *
from .image import *
from .network import *
from .provider import *
from .remote_image import *
from .secret import *
from .service import *
from .service_config import *
from .utils import *
from .volume import *
from ._inputs import *
from . import outputs

# Make subpackages available:
from . import (
    config,
)



# coding=utf-8
# *** WARNING: this file was generated by the Pulumi Terraform Bridge (tfgen) Tool. ***
# *** Do not edit by hand unless you're certain you know what you are doing! ***

import warnings
import pulumi
import pulumi.runtime
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
from . import _utilities, _tables
from . import outputs

__all__ = [
    'ContainerCapabilities',
    'ContainerDevice',
    'ContainerHealthcheck',
    'ContainerHost',
    'ContainerLabel',
    'ContainerMount',
    'ContainerMountBindOptions',
    'ContainerMountTmpfsOptions',
    'ContainerMountVolumeOptions',
    'ContainerMountVolumeOptionsLabel',
    'ContainerNetworkData',
    'ContainerNetworksAdvanced',
    'ContainerPort',
    'ContainerUlimit',
    'ContainerUpload',
    'ContainerVolume',
    'NetworkIpamConfig',
    'NetworkLabel',
    'SecretLabel',
    'ServiceAuth',
    'ServiceConvergeConfig',
    'ServiceEndpointSpec',
    'ServiceEndpointSpecPort',
    'ServiceLabel',
    'ServiceMode',
    'ServiceModeReplicated',
    'ServiceRollbackConfig',
    'ServiceTaskSpec',
    'ServiceTaskSpecContainerSpec',
    'ServiceTaskSpecContainerSpecConfig',
    'ServiceTaskSpecContainerSpecDnsConfig',
    'ServiceTaskSpecContainerSpecHealthcheck',
    'ServiceTaskSpecContainerSpecHost',
    'ServiceTaskSpecContainerSpecLabel',
    'ServiceTaskSpecContainerSpecMount',
    'ServiceTaskSpecContainerSpecMountBindOptions',
    'ServiceTaskSpecContainerSpecMountTmpfsOptions',
    'ServiceTaskSpecContainerSpecMountVolumeOptions',
    'ServiceTaskSpecContainerSpecMountVolumeOptionsLabel',
    'ServiceTaskSpecContainerSpecPrivileges',
    'ServiceTaskSpecContainerSpecPrivilegesCredentialSpec',
    'ServiceTaskSpecContainerSpecPrivilegesSeLinuxContext',
    'ServiceTaskSpecContainerSpecSecret',
    'ServiceTaskSpecLogDriver',
    'ServiceTaskSpecPlacement',
    'ServiceTaskSpecPlacementPlatform',
    'ServiceTaskSpecResources',
    'ServiceTaskSpecResourcesLimits',
    'ServiceTaskSpecResourcesLimitsGenericResources',
    'ServiceTaskSpecResourcesReservation',
    'ServiceTaskSpecResourcesReservationGenericResources',
    'ServiceTaskSpecRestartPolicy',
    'ServiceUpdateConfig',
    'VolumeLabel',
    'GetNetworkIpamConfigResult',
    'ProviderRegistryAuth',
]

@pulumi.output_type
class ContainerCapabilities(dict):
    def __init__(__self__, *,
                 adds: Optional[List[str]] = None,
                 drops: Optional[List[str]] = None):
        """
        :param List[str] adds: list of linux capabilities to add.
        :param List[str] drops: list of linux capabilities to drop.
        """
        if adds is not None:
            pulumi.set(__self__, "adds", adds)
        if drops is not None:
            pulumi.set(__self__, "drops", drops)

    @property
    @pulumi.getter
    def adds(self) -> Optional[List[str]]:
        """
        list of linux capabilities to add.
        """
        return pulumi.get(self, "adds")

    @property
    @pulumi.getter
    def drops(self) -> Optional[List[str]]:
        """
        list of linux capabilities to drop.
        """
        return pulumi.get(self, "drops")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerDevice(dict):
    def __init__(__self__, *,
                 host_path: str,
                 container_path: Optional[str] = None,
                 permissions: Optional[str] = None):
        """
        :param str host_path: The path on the host where the device
               is located.
        :param str container_path: The path in the container where the
               device will be binded.
        :param str permissions: The cgroup permissions given to the
               container to access the device.
               Defaults to `rwm`.
        """
        pulumi.set(__self__, "host_path", host_path)
        if container_path is not None:
            pulumi.set(__self__, "container_path", container_path)
        if permissions is not None:
            pulumi.set(__self__, "permissions", permissions)

    @property
    @pulumi.getter(name="hostPath")
    def host_path(self) -> str:
        """
        The path on the host where the device
        is located.
        """
        return pulumi.get(self, "host_path")

    @property
    @pulumi.getter(name="containerPath")
    def container_path(self) -> Optional[str]:
        """
        The path in the container where the
        device will be binded.
        """
        return pulumi.get(self, "container_path")

    @property
    @pulumi.getter
    def permissions(self) -> Optional[str]:
        """
        The cgroup permissions given to the
        container to access the device.
        Defaults to `rwm`.
        """
        return pulumi.get(self, "permissions")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerHealthcheck(dict):
    def __init__(__self__, *,
                 tests: List[str],
                 interval: Optional[str] = None,
                 retries: Optional[float] = None,
                 start_period: Optional[str] = None,
                 timeout: Optional[str] = None):
        """
        :param List[str] tests: Command to run to check health. For example, to run `curl -f http://localhost/health` set the
               command to be `["CMD", "curl", "-f", "http://localhost/health"]`.
        :param str interval: Time between running the check `(ms|s|m|h)`. Default: `0s`.
        :param float retries: Consecutive failures needed to report unhealthy. Default: `0`.
        :param str start_period: Start period for the container to initialize before counting retries towards unstable `(ms|s|m|h)`. Default: `0s`.
        :param str timeout: Maximum time to allow one check to run `(ms|s|m|h)`. Default: `0s`.
        """
        pulumi.set(__self__, "tests", tests)
        if interval is not None:
            pulumi.set(__self__, "interval", interval)
        if retries is not None:
            pulumi.set(__self__, "retries", retries)
        if start_period is not None:
            pulumi.set(__self__, "start_period", start_period)
        if timeout is not None:
            pulumi.set(__self__, "timeout", timeout)

    @property
    @pulumi.getter
    def tests(self) -> List[str]:
        """
        Command to run to check health. For example, to run `curl -f http://localhost/health` set the
        command to be `["CMD", "curl", "-f", "http://localhost/health"]`.
        """
        return pulumi.get(self, "tests")

    @property
    @pulumi.getter
    def interval(self) -> Optional[str]:
        """
        Time between running the check `(ms|s|m|h)`. Default: `0s`.
        """
        return pulumi.get(self, "interval")

    @property
    @pulumi.getter
    def retries(self) -> Optional[float]:
        """
        Consecutive failures needed to report unhealthy. Default: `0`.
        """
        return pulumi.get(self, "retries")

    @property
    @pulumi.getter(name="startPeriod")
    def start_period(self) -> Optional[str]:
        """
        Start period for the container to initialize before counting retries towards unstable `(ms|s|m|h)`. Default: `0s`.
        """
        return pulumi.get(self, "start_period")

    @property
    @pulumi.getter
    def timeout(self) -> Optional[str]:
        """
        Maximum time to allow one check to run `(ms|s|m|h)`. Default: `0s`.
        """
        return pulumi.get(self, "timeout")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerHost(dict):
    def __init__(__self__, *,
                 host: str,
                 ip: str):
        """
        :param str host: Hostname to add.
        :param str ip: IP address this hostname should resolve to.
        """
        pulumi.set(__self__, "host", host)
        pulumi.set(__self__, "ip", ip)

    @property
    @pulumi.getter
    def host(self) -> str:
        """
        Hostname to add.
        """
        return pulumi.get(self, "host")

    @property
    @pulumi.getter
    def ip(self) -> str:
        """
        IP address this hostname should resolve to.
        """
        return pulumi.get(self, "ip")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        """
        :param str label: Name of the label
        :param str value: Value of the label
        """
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        """
        Name of the label
        """
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        """
        Value of the label
        """
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerMount(dict):
    def __init__(__self__, *,
                 target: str,
                 type: str,
                 bind_options: Optional['outputs.ContainerMountBindOptions'] = None,
                 read_only: Optional[bool] = None,
                 source: Optional[str] = None,
                 tmpfs_options: Optional['outputs.ContainerMountTmpfsOptions'] = None,
                 volume_options: Optional['outputs.ContainerMountVolumeOptions'] = None):
        """
        :param str target: The container path.
        :param str type: The mount type: valid values are `bind|volume|tmpfs`.
        :param 'ContainerMountBindOptionsArgs' bind_options: Optional configuration for the `bind` type.
        :param bool read_only: If true, this volume will be readonly.
               Defaults to false.
        :param str source: The mount source (e.g., a volume name, a host path)
        :param 'ContainerMountTmpfsOptionsArgs' tmpfs_options: Optional configuration for the `tmpf` type.
        :param 'ContainerMountVolumeOptionsArgs' volume_options: Optional configuration for the `volume` type.
        """
        pulumi.set(__self__, "target", target)
        pulumi.set(__self__, "type", type)
        if bind_options is not None:
            pulumi.set(__self__, "bind_options", bind_options)
        if read_only is not None:
            pulumi.set(__self__, "read_only", read_only)
        if source is not None:
            pulumi.set(__self__, "source", source)
        if tmpfs_options is not None:
            pulumi.set(__self__, "tmpfs_options", tmpfs_options)
        if volume_options is not None:
            pulumi.set(__self__, "volume_options", volume_options)

    @property
    @pulumi.getter
    def target(self) -> str:
        """
        The container path.
        """
        return pulumi.get(self, "target")

    @property
    @pulumi.getter
    def type(self) -> str:
        """
        The mount type: valid values are `bind|volume|tmpfs`.
        """
        return pulumi.get(self, "type")

    @property
    @pulumi.getter(name="bindOptions")
    def bind_options(self) -> Optional['outputs.ContainerMountBindOptions']:
        """
        Optional configuration for the `bind` type.
        """
        return pulumi.get(self, "bind_options")

    @property
    @pulumi.getter(name="readOnly")
    def read_only(self) -> Optional[bool]:
        """
        If true, this volume will be readonly.
        Defaults to false.
        """
        return pulumi.get(self, "read_only")

    @property
    @pulumi.getter
    def source(self) -> Optional[str]:
        """
        The mount source (e.g., a volume name, a host path)
        """
        return pulumi.get(self, "source")

    @property
    @pulumi.getter(name="tmpfsOptions")
    def tmpfs_options(self) -> Optional['outputs.ContainerMountTmpfsOptions']:
        """
        Optional configuration for the `tmpf` type.
        """
        return pulumi.get(self, "tmpfs_options")

    @property
    @pulumi.getter(name="volumeOptions")
    def volume_options(self) -> Optional['outputs.ContainerMountVolumeOptions']:
        """
        Optional configuration for the `volume` type.
        """
        return pulumi.get(self, "volume_options")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerMountBindOptions(dict):
    def __init__(__self__, *,
                 propagation: Optional[str] = None):
        """
        :param str propagation: A propagation mode with the value.
        """
        if propagation is not None:
            pulumi.set(__self__, "propagation", propagation)

    @property
    @pulumi.getter
    def propagation(self) -> Optional[str]:
        """
        A propagation mode with the value.
        """
        return pulumi.get(self, "propagation")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerMountTmpfsOptions(dict):
    def __init__(__self__, *,
                 mode: Optional[float] = None,
                 size_bytes: Optional[float] = None):
        """
        :param float mode: The permission mode for the tmpfs mount in an integer.
        :param float size_bytes: The size for the tmpfs mount in bytes.
        """
        if mode is not None:
            pulumi.set(__self__, "mode", mode)
        if size_bytes is not None:
            pulumi.set(__self__, "size_bytes", size_bytes)

    @property
    @pulumi.getter
    def mode(self) -> Optional[float]:
        """
        The permission mode for the tmpfs mount in an integer.
        """
        return pulumi.get(self, "mode")

    @property
    @pulumi.getter(name="sizeBytes")
    def size_bytes(self) -> Optional[float]:
        """
        The size for the tmpfs mount in bytes.
        """
        return pulumi.get(self, "size_bytes")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerMountVolumeOptions(dict):
    def __init__(__self__, *,
                 driver_name: Optional[str] = None,
                 driver_options: Optional[Mapping[str, str]] = None,
                 labels: Optional[List['outputs.ContainerMountVolumeOptionsLabel']] = None,
                 no_copy: Optional[bool] = None):
        """
        :param Mapping[str, str] driver_options: Options for the driver.
        :param List['ContainerMountVolumeOptionsLabelArgs'] labels: Adding labels.
        :param bool no_copy: Whether to populate volume with data from the target.
        """
        if driver_name is not None:
            pulumi.set(__self__, "driver_name", driver_name)
        if driver_options is not None:
            pulumi.set(__self__, "driver_options", driver_options)
        if labels is not None:
            pulumi.set(__self__, "labels", labels)
        if no_copy is not None:
            pulumi.set(__self__, "no_copy", no_copy)

    @property
    @pulumi.getter(name="driverName")
    def driver_name(self) -> Optional[str]:
        return pulumi.get(self, "driver_name")

    @property
    @pulumi.getter(name="driverOptions")
    def driver_options(self) -> Optional[Mapping[str, str]]:
        """
        Options for the driver.
        """
        return pulumi.get(self, "driver_options")

    @property
    @pulumi.getter
    def labels(self) -> Optional[List['outputs.ContainerMountVolumeOptionsLabel']]:
        """
        Adding labels.
        """
        return pulumi.get(self, "labels")

    @property
    @pulumi.getter(name="noCopy")
    def no_copy(self) -> Optional[bool]:
        """
        Whether to populate volume with data from the target.
        """
        return pulumi.get(self, "no_copy")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerMountVolumeOptionsLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        """
        :param str label: Name of the label
        :param str value: Value of the label
        """
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        """
        Name of the label
        """
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        """
        Value of the label
        """
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerNetworkData(dict):
    def __init__(__self__, *,
                 gateway: Optional[str] = None,
                 ip_address: Optional[str] = None,
                 ip_prefix_length: Optional[float] = None,
                 network_name: Optional[str] = None):
        """
        :param str gateway: *Deprecated:* Use `network_data` instead. The network gateway of the container as read from its
               NetworkSettings.
        :param str ip_address: *Deprecated:* Use `network_data` instead. The IP address of the container's first network it.
        :param float ip_prefix_length: *Deprecated:* Use `network_data` instead. The IP prefix length of the container as read from its
               NetworkSettings.
        """
        if gateway is not None:
            pulumi.set(__self__, "gateway", gateway)
        if ip_address is not None:
            pulumi.set(__self__, "ip_address", ip_address)
        if ip_prefix_length is not None:
            pulumi.set(__self__, "ip_prefix_length", ip_prefix_length)
        if network_name is not None:
            pulumi.set(__self__, "network_name", network_name)

    @property
    @pulumi.getter
    def gateway(self) -> Optional[str]:
        """
        *Deprecated:* Use `network_data` instead. The network gateway of the container as read from its
        NetworkSettings.
        """
        return pulumi.get(self, "gateway")

    @property
    @pulumi.getter(name="ipAddress")
    def ip_address(self) -> Optional[str]:
        """
        *Deprecated:* Use `network_data` instead. The IP address of the container's first network it.
        """
        return pulumi.get(self, "ip_address")

    @property
    @pulumi.getter(name="ipPrefixLength")
    def ip_prefix_length(self) -> Optional[float]:
        """
        *Deprecated:* Use `network_data` instead. The IP prefix length of the container as read from its
        NetworkSettings.
        """
        return pulumi.get(self, "ip_prefix_length")

    @property
    @pulumi.getter(name="networkName")
    def network_name(self) -> Optional[str]:
        return pulumi.get(self, "network_name")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerNetworksAdvanced(dict):
    def __init__(__self__, *,
                 name: str,
                 aliases: Optional[List[str]] = None,
                 ipv4_address: Optional[str] = None,
                 ipv6_address: Optional[str] = None):
        """
        :param str name: The name of the network.
        :param List[str] aliases: The network aliases of the container in the specific network.
        :param str ipv4_address: The IPV4 address of the container in the specific network.
        :param str ipv6_address: The IPV6 address of the container in the specific network.
        """
        pulumi.set(__self__, "name", name)
        if aliases is not None:
            pulumi.set(__self__, "aliases", aliases)
        if ipv4_address is not None:
            pulumi.set(__self__, "ipv4_address", ipv4_address)
        if ipv6_address is not None:
            pulumi.set(__self__, "ipv6_address", ipv6_address)

    @property
    @pulumi.getter
    def name(self) -> str:
        """
        The name of the network.
        """
        return pulumi.get(self, "name")

    @property
    @pulumi.getter
    def aliases(self) -> Optional[List[str]]:
        """
        The network aliases of the container in the specific network.
        """
        return pulumi.get(self, "aliases")

    @property
    @pulumi.getter(name="ipv4Address")
    def ipv4_address(self) -> Optional[str]:
        """
        The IPV4 address of the container in the specific network.
        """
        return pulumi.get(self, "ipv4_address")

    @property
    @pulumi.getter(name="ipv6Address")
    def ipv6_address(self) -> Optional[str]:
        """
        The IPV6 address of the container in the specific network.
        """
        return pulumi.get(self, "ipv6_address")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerPort(dict):
    def __init__(__self__, *,
                 internal: float,
                 external: Optional[float] = None,
                 ip: Optional[str] = None,
                 protocol: Optional[str] = None):
        """
        :param float internal: Port within the container.
        :param float external: Port exposed out of the container. If not given a free random port `>= 32768` will be used.
        :param str ip: IP address this hostname should resolve to.
        :param str protocol: Protocol that can be used over this port,
               defaults to `tcp`.
        """
        pulumi.set(__self__, "internal", internal)
        if external is not None:
            pulumi.set(__self__, "external", external)
        if ip is not None:
            pulumi.set(__self__, "ip", ip)
        if protocol is not None:
            pulumi.set(__self__, "protocol", protocol)

    @property
    @pulumi.getter
    def internal(self) -> float:
        """
        Port within the container.
        """
        return pulumi.get(self, "internal")

    @property
    @pulumi.getter
    def external(self) -> Optional[float]:
        """
        Port exposed out of the container. If not given a free random port `>= 32768` will be used.
        """
        return pulumi.get(self, "external")

    @property
    @pulumi.getter
    def ip(self) -> Optional[str]:
        """
        IP address this hostname should resolve to.
        """
        return pulumi.get(self, "ip")

    @property
    @pulumi.getter
    def protocol(self) -> Optional[str]:
        """
        Protocol that can be used over this port,
        defaults to `tcp`.
        """
        return pulumi.get(self, "protocol")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerUlimit(dict):
    def __init__(__self__, *,
                 hard: float,
                 name: str,
                 soft: float):
        pulumi.set(__self__, "hard", hard)
        pulumi.set(__self__, "name", name)
        pulumi.set(__self__, "soft", soft)

    @property
    @pulumi.getter
    def hard(self) -> float:
        return pulumi.get(self, "hard")

    @property
    @pulumi.getter
    def name(self) -> str:
        return pulumi.get(self, "name")

    @property
    @pulumi.getter
    def soft(self) -> float:
        return pulumi.get(self, "soft")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerUpload(dict):
    def __init__(__self__, *,
                 file: str,
                 content: Optional[str] = None,
                 content_base64: Optional[str] = None,
                 executable: Optional[bool] = None,
                 source: Optional[str] = None,
                 source_hash: Optional[str] = None):
        """
        :param str file: path to a file in the container.
        :param str content: Literal string value to use as the object content, which will be uploaded as UTF-8-encoded text.
        :param bool executable: If true, the file will be uploaded with user
               executable permission.
               Defaults to false.
        :param str source: A filename that references a file which will be uploaded as the object content. This allows for large file uploads that do not get stored in state.
        :param str source_hash: If using `source`, this will force an update if the file content has updated but the filename has not.
        """
        pulumi.set(__self__, "file", file)
        if content is not None:
            pulumi.set(__self__, "content", content)
        if content_base64 is not None:
            pulumi.set(__self__, "content_base64", content_base64)
        if executable is not None:
            pulumi.set(__self__, "executable", executable)
        if source is not None:
            pulumi.set(__self__, "source", source)
        if source_hash is not None:
            pulumi.set(__self__, "source_hash", source_hash)

    @property
    @pulumi.getter
    def file(self) -> str:
        """
        path to a file in the container.
        """
        return pulumi.get(self, "file")

    @property
    @pulumi.getter
    def content(self) -> Optional[str]:
        """
        Literal string value to use as the object content, which will be uploaded as UTF-8-encoded text.
        """
        return pulumi.get(self, "content")

    @property
    @pulumi.getter(name="contentBase64")
    def content_base64(self) -> Optional[str]:
        return pulumi.get(self, "content_base64")

    @property
    @pulumi.getter
    def executable(self) -> Optional[bool]:
        """
        If true, the file will be uploaded with user
        executable permission.
        Defaults to false.
        """
        return pulumi.get(self, "executable")

    @property
    @pulumi.getter
    def source(self) -> Optional[str]:
        """
        A filename that references a file which will be uploaded as the object content. This allows for large file uploads that do not get stored in state.
        """
        return pulumi.get(self, "source")

    @property
    @pulumi.getter(name="sourceHash")
    def source_hash(self) -> Optional[str]:
        """
        If using `source`, this will force an update if the file content has updated but the filename has not.
        """
        return pulumi.get(self, "source_hash")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ContainerVolume(dict):
    def __init__(__self__, *,
                 container_path: Optional[str] = None,
                 from_container: Optional[str] = None,
                 host_path: Optional[str] = None,
                 read_only: Optional[bool] = None,
                 volume_name: Optional[str] = None):
        """
        :param str container_path: The path in the container where the
               device will be binded.
        :param str from_container: The container where the volume is
               coming from.
        :param str host_path: The path on the host where the device
               is located.
        :param bool read_only: If true, this volume will be readonly.
               Defaults to false.
        :param str volume_name: The name of the docker volume which
               should be mounted.
        """
        if container_path is not None:
            pulumi.set(__self__, "container_path", container_path)
        if from_container is not None:
            pulumi.set(__self__, "from_container", from_container)
        if host_path is not None:
            pulumi.set(__self__, "host_path", host_path)
        if read_only is not None:
            pulumi.set(__self__, "read_only", read_only)
        if volume_name is not None:
            pulumi.set(__self__, "volume_name", volume_name)

    @property
    @pulumi.getter(name="containerPath")
    def container_path(self) -> Optional[str]:
        """
        The path in the container where the
        device will be binded.
        """
        return pulumi.get(self, "container_path")

    @property
    @pulumi.getter(name="fromContainer")
    def from_container(self) -> Optional[str]:
        """
        The container where the volume is
        coming from.
        """
        return pulumi.get(self, "from_container")

    @property
    @pulumi.getter(name="hostPath")
    def host_path(self) -> Optional[str]:
        """
        The path on the host where the device
        is located.
        """
        return pulumi.get(self, "host_path")

    @property
    @pulumi.getter(name="readOnly")
    def read_only(self) -> Optional[bool]:
        """
        If true, this volume will be readonly.
        Defaults to false.
        """
        return pulumi.get(self, "read_only")

    @property
    @pulumi.getter(name="volumeName")
    def volume_name(self) -> Optional[str]:
        """
        The name of the docker volume which
        should be mounted.
        """
        return pulumi.get(self, "volume_name")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class NetworkIpamConfig(dict):
    def __init__(__self__, *,
                 aux_address: Optional[Mapping[str, Any]] = None,
                 gateway: Optional[str] = None,
                 ip_range: Optional[str] = None,
                 subnet: Optional[str] = None):
        if aux_address is not None:
            pulumi.set(__self__, "aux_address", aux_address)
        if gateway is not None:
            pulumi.set(__self__, "gateway", gateway)
        if ip_range is not None:
            pulumi.set(__self__, "ip_range", ip_range)
        if subnet is not None:
            pulumi.set(__self__, "subnet", subnet)

    @property
    @pulumi.getter(name="auxAddress")
    def aux_address(self) -> Optional[Mapping[str, Any]]:
        return pulumi.get(self, "aux_address")

    @property
    @pulumi.getter
    def gateway(self) -> Optional[str]:
        return pulumi.get(self, "gateway")

    @property
    @pulumi.getter(name="ipRange")
    def ip_range(self) -> Optional[str]:
        return pulumi.get(self, "ip_range")

    @property
    @pulumi.getter
    def subnet(self) -> Optional[str]:
        return pulumi.get(self, "subnet")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class NetworkLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        """
        :param str label: Name of the label
        :param str value: Value of the label
        """
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        """
        Name of the label
        """
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        """
        Value of the label
        """
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class SecretLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        """
        :param str label: Name of the label
        :param str value: Value of the label
        """
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        """
        Name of the label
        """
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        """
        Value of the label
        """
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceAuth(dict):
    def __init__(__self__, *,
                 server_address: str,
                 password: Optional[str] = None,
                 username: Optional[str] = None):
        """
        :param str server_address: The address of the registry server
        :param str password: The password to use for authenticating to the registry. If this is blank, the `DOCKER_REGISTRY_PASS` is also be checked.
        :param str username: The username to use for authenticating to the registry. If this is blank, the `DOCKER_REGISTRY_USER` is also be checked.
        """
        pulumi.set(__self__, "server_address", server_address)
        if password is not None:
            pulumi.set(__self__, "password", password)
        if username is not None:
            pulumi.set(__self__, "username", username)

    @property
    @pulumi.getter(name="serverAddress")
    def server_address(self) -> str:
        """
        The address of the registry server
        """
        return pulumi.get(self, "server_address")

    @property
    @pulumi.getter
    def password(self) -> Optional[str]:
        """
        The password to use for authenticating to the registry. If this is blank, the `DOCKER_REGISTRY_PASS` is also be checked.
        """
        return pulumi.get(self, "password")

    @property
    @pulumi.getter
    def username(self) -> Optional[str]:
        """
        The username to use for authenticating to the registry. If this is blank, the `DOCKER_REGISTRY_USER` is also be checked.
        """
        return pulumi.get(self, "username")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceConvergeConfig(dict):
    def __init__(__self__, *,
                 delay: Optional[str] = None,
                 timeout: Optional[str] = None):
        """
        :param str delay: Time between each the check to check docker endpoint `(ms|s|m|h)`. For example, to check if
               all tasks are up when a service is created, or to check if all tasks are successfully updated on an update. Default: `7s`.
        :param str timeout: The timeout of the service to reach the desired state `(s|m)`. Default: `3m`.
        """
        if delay is not None:
            pulumi.set(__self__, "delay", delay)
        if timeout is not None:
            pulumi.set(__self__, "timeout", timeout)

    @property
    @pulumi.getter
    def delay(self) -> Optional[str]:
        """
        Time between each the check to check docker endpoint `(ms|s|m|h)`. For example, to check if
        all tasks are up when a service is created, or to check if all tasks are successfully updated on an update. Default: `7s`.
        """
        return pulumi.get(self, "delay")

    @property
    @pulumi.getter
    def timeout(self) -> Optional[str]:
        """
        The timeout of the service to reach the desired state `(s|m)`. Default: `3m`.
        """
        return pulumi.get(self, "timeout")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceEndpointSpec(dict):
    def __init__(__self__, *,
                 mode: Optional[str] = None,
                 ports: Optional[List['outputs.ServiceEndpointSpecPort']] = None):
        """
        :param str mode: The mode of resolution to use for internal load balancing between tasks. `(vip|dnsrr)`. Default: `vip`.
        :param List['ServiceEndpointSpecPortArgs'] ports: See Ports below for details.
        """
        if mode is not None:
            pulumi.set(__self__, "mode", mode)
        if ports is not None:
            pulumi.set(__self__, "ports", ports)

    @property
    @pulumi.getter
    def mode(self) -> Optional[str]:
        """
        The mode of resolution to use for internal load balancing between tasks. `(vip|dnsrr)`. Default: `vip`.
        """
        return pulumi.get(self, "mode")

    @property
    @pulumi.getter
    def ports(self) -> Optional[List['outputs.ServiceEndpointSpecPort']]:
        """
        See Ports below for details.
        """
        return pulumi.get(self, "ports")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceEndpointSpecPort(dict):
    def __init__(__self__, *,
                 target_port: float,
                 name: Optional[str] = None,
                 protocol: Optional[str] = None,
                 publish_mode: Optional[str] = None,
                 published_port: Optional[float] = None):
        """
        :param float target_port: Port inside the container.
        :param str name: The name of the Docker service.
        :param str protocol: Protocol that can be used over this port: `tcp|udp|sctp`. Default: `tcp`.
        :param str publish_mode: Represents the mode in which the port is to be published: `ingress|host`
        :param float published_port: The port on the swarm hosts. If not set the value of `target_port` will be used.
        """
        pulumi.set(__self__, "target_port", target_port)
        if name is not None:
            pulumi.set(__self__, "name", name)
        if protocol is not None:
            pulumi.set(__self__, "protocol", protocol)
        if publish_mode is not None:
            pulumi.set(__self__, "publish_mode", publish_mode)
        if published_port is not None:
            pulumi.set(__self__, "published_port", published_port)

    @property
    @pulumi.getter(name="targetPort")
    def target_port(self) -> float:
        """
        Port inside the container.
        """
        return pulumi.get(self, "target_port")

    @property
    @pulumi.getter
    def name(self) -> Optional[str]:
        """
        The name of the Docker service.
        """
        return pulumi.get(self, "name")

    @property
    @pulumi.getter
    def protocol(self) -> Optional[str]:
        """
        Protocol that can be used over this port: `tcp|udp|sctp`. Default: `tcp`.
        """
        return pulumi.get(self, "protocol")

    @property
    @pulumi.getter(name="publishMode")
    def publish_mode(self) -> Optional[str]:
        """
        Represents the mode in which the port is to be published: `ingress|host`
        """
        return pulumi.get(self, "publish_mode")

    @property
    @pulumi.getter(name="publishedPort")
    def published_port(self) -> Optional[float]:
        """
        The port on the swarm hosts. If not set the value of `target_port` will be used.
        """
        return pulumi.get(self, "published_port")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        """
        :param str label: Name of the label
        :param str value: Value of the label
        """
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        """
        Name of the label
        """
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        """
        Value of the label
        """
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceMode(dict):
    def __init__(__self__, *,
                 global_: Optional[bool] = None,
                 replicated: Optional['outputs.ServiceModeReplicated'] = None):
        """
        :param bool global_: set it to `true` to run the service in the global mode
        :param 'ServiceModeReplicatedArgs' replicated: , which contains atm only the amount of `replicas`
        """
        if global_ is not None:
            pulumi.set(__self__, "global_", global_)
        if replicated is not None:
            pulumi.set(__self__, "replicated", replicated)

    @property
    @pulumi.getter(name="global")
    def global_(self) -> Optional[bool]:
        """
        set it to `true` to run the service in the global mode
        """
        return pulumi.get(self, "global_")

    @property
    @pulumi.getter
    def replicated(self) -> Optional['outputs.ServiceModeReplicated']:
        """
        , which contains atm only the amount of `replicas`
        """
        return pulumi.get(self, "replicated")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceModeReplicated(dict):
    def __init__(__self__, *,
                 replicas: Optional[float] = None):
        if replicas is not None:
            pulumi.set(__self__, "replicas", replicas)

    @property
    @pulumi.getter
    def replicas(self) -> Optional[float]:
        return pulumi.get(self, "replicas")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceRollbackConfig(dict):
    def __init__(__self__, *,
                 delay: Optional[str] = None,
                 failure_action: Optional[str] = None,
                 max_failure_ratio: Optional[str] = None,
                 monitor: Optional[str] = None,
                 order: Optional[str] = None,
                 parallelism: Optional[float] = None):
        """
        :param str delay: Delay between restart attempts `(ms|s|m|h)`
               all tasks are up when a service is created, or to check if all tasks are successfully updated on an update. Default: `7s`.
        :param str failure_action: Action on update failure: `pause|continue|rollback`.
        :param str max_failure_ratio: The failure rate to tolerate during an update as `float`. **Important:** the `float`need to be wrapped in a `string` to avoid internal
               casting and precision errors.
        :param str monitor: Duration after each task update to monitor for failure `(ns|us|ms|s|m|h)`
        :param str order: Update order either 'stop-first' or 'start-first'.
        :param float parallelism: The maximum number of tasks to be updated in one iteration simultaneously (0 to update all at once).
        """
        if delay is not None:
            pulumi.set(__self__, "delay", delay)
        if failure_action is not None:
            pulumi.set(__self__, "failure_action", failure_action)
        if max_failure_ratio is not None:
            pulumi.set(__self__, "max_failure_ratio", max_failure_ratio)
        if monitor is not None:
            pulumi.set(__self__, "monitor", monitor)
        if order is not None:
            pulumi.set(__self__, "order", order)
        if parallelism is not None:
            pulumi.set(__self__, "parallelism", parallelism)

    @property
    @pulumi.getter
    def delay(self) -> Optional[str]:
        """
        Delay between restart attempts `(ms|s|m|h)`
        all tasks are up when a service is created, or to check if all tasks are successfully updated on an update. Default: `7s`.
        """
        return pulumi.get(self, "delay")

    @property
    @pulumi.getter(name="failureAction")
    def failure_action(self) -> Optional[str]:
        """
        Action on update failure: `pause|continue|rollback`.
        """
        return pulumi.get(self, "failure_action")

    @property
    @pulumi.getter(name="maxFailureRatio")
    def max_failure_ratio(self) -> Optional[str]:
        """
        The failure rate to tolerate during an update as `float`. **Important:** the `float`need to be wrapped in a `string` to avoid internal
        casting and precision errors.
        """
        return pulumi.get(self, "max_failure_ratio")

    @property
    @pulumi.getter
    def monitor(self) -> Optional[str]:
        """
        Duration after each task update to monitor for failure `(ns|us|ms|s|m|h)`
        """
        return pulumi.get(self, "monitor")

    @property
    @pulumi.getter
    def order(self) -> Optional[str]:
        """
        Update order either 'stop-first' or 'start-first'.
        """
        return pulumi.get(self, "order")

    @property
    @pulumi.getter
    def parallelism(self) -> Optional[float]:
        """
        The maximum number of tasks to be updated in one iteration simultaneously (0 to update all at once).
        """
        return pulumi.get(self, "parallelism")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpec(dict):
    def __init__(__self__, *,
                 container_spec: 'outputs.ServiceTaskSpecContainerSpec',
                 force_update: Optional[float] = None,
                 log_driver: Optional['outputs.ServiceTaskSpecLogDriver'] = None,
                 networks: Optional[List[str]] = None,
                 placement: Optional['outputs.ServiceTaskSpecPlacement'] = None,
                 resources: Optional['outputs.ServiceTaskSpecResources'] = None,
                 restart_policy: Optional['outputs.ServiceTaskSpecRestartPolicy'] = None,
                 runtime: Optional[str] = None):
        """
        :param 'ServiceTaskSpecContainerSpecArgs' container_spec: See ContainerSpec below for details.
        :param float force_update: A counter that triggers an update even if no relevant parameters have been changed. See [Docker Spec](https://github.com/docker/swarmkit/blob/master/api/specs.proto#L126).
        :param 'ServiceTaskSpecLogDriverArgs' log_driver: See Log Driver below for details.
        :param List[str] networks: Ids of the networks in which the container will be put in.
        :param 'ServiceTaskSpecPlacementArgs' placement: See Placement below for details.
        :param 'ServiceTaskSpecResourcesArgs' resources: See Resources below for details.
        :param 'ServiceTaskSpecRestartPolicyArgs' restart_policy: See Restart Policy below for details.
        :param str runtime: Runtime is the type of runtime specified for the task executor. See [Docker Runtime](https://github.com/moby/moby/blob/master/api/types/swarm/runtime.go).
        """
        pulumi.set(__self__, "container_spec", container_spec)
        if force_update is not None:
            pulumi.set(__self__, "force_update", force_update)
        if log_driver is not None:
            pulumi.set(__self__, "log_driver", log_driver)
        if networks is not None:
            pulumi.set(__self__, "networks", networks)
        if placement is not None:
            pulumi.set(__self__, "placement", placement)
        if resources is not None:
            pulumi.set(__self__, "resources", resources)
        if restart_policy is not None:
            pulumi.set(__self__, "restart_policy", restart_policy)
        if runtime is not None:
            pulumi.set(__self__, "runtime", runtime)

    @property
    @pulumi.getter(name="containerSpec")
    def container_spec(self) -> 'outputs.ServiceTaskSpecContainerSpec':
        """
        See ContainerSpec below for details.
        """
        return pulumi.get(self, "container_spec")

    @property
    @pulumi.getter(name="forceUpdate")
    def force_update(self) -> Optional[float]:
        """
        A counter that triggers an update even if no relevant parameters have been changed. See [Docker Spec](https://github.com/docker/swarmkit/blob/master/api/specs.proto#L126).
        """
        return pulumi.get(self, "force_update")

    @property
    @pulumi.getter(name="logDriver")
    def log_driver(self) -> Optional['outputs.ServiceTaskSpecLogDriver']:
        """
        See Log Driver below for details.
        """
        return pulumi.get(self, "log_driver")

    @property
    @pulumi.getter
    def networks(self) -> Optional[List[str]]:
        """
        Ids of the networks in which the container will be put in.
        """
        return pulumi.get(self, "networks")

    @property
    @pulumi.getter
    def placement(self) -> Optional['outputs.ServiceTaskSpecPlacement']:
        """
        See Placement below for details.
        """
        return pulumi.get(self, "placement")

    @property
    @pulumi.getter
    def resources(self) -> Optional['outputs.ServiceTaskSpecResources']:
        """
        See Resources below for details.
        """
        return pulumi.get(self, "resources")

    @property
    @pulumi.getter(name="restartPolicy")
    def restart_policy(self) -> Optional['outputs.ServiceTaskSpecRestartPolicy']:
        """
        See Restart Policy below for details.
        """
        return pulumi.get(self, "restart_policy")

    @property
    @pulumi.getter
    def runtime(self) -> Optional[str]:
        """
        Runtime is the type of runtime specified for the task executor. See [Docker Runtime](https://github.com/moby/moby/blob/master/api/types/swarm/runtime.go).
        """
        return pulumi.get(self, "runtime")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpec(dict):
    def __init__(__self__, *,
                 image: str,
                 args: Optional[List[str]] = None,
                 commands: Optional[List[str]] = None,
                 configs: Optional[List['outputs.ServiceTaskSpecContainerSpecConfig']] = None,
                 dir: Optional[str] = None,
                 dns_config: Optional['outputs.ServiceTaskSpecContainerSpecDnsConfig'] = None,
                 env: Optional[Mapping[str, str]] = None,
                 groups: Optional[List[str]] = None,
                 healthcheck: Optional['outputs.ServiceTaskSpecContainerSpecHealthcheck'] = None,
                 hostname: Optional[str] = None,
                 hosts: Optional[List['outputs.ServiceTaskSpecContainerSpecHost']] = None,
                 isolation: Optional[str] = None,
                 labels: Optional[List['outputs.ServiceTaskSpecContainerSpecLabel']] = None,
                 mounts: Optional[List['outputs.ServiceTaskSpecContainerSpecMount']] = None,
                 privileges: Optional['outputs.ServiceTaskSpecContainerSpecPrivileges'] = None,
                 read_only: Optional[bool] = None,
                 secrets: Optional[List['outputs.ServiceTaskSpecContainerSpecSecret']] = None,
                 stop_grace_period: Optional[str] = None,
                 stop_signal: Optional[str] = None,
                 user: Optional[str] = None):
        """
        :param str image: The image used to create the Docker service.
        :param List[str] args: Arguments to the command.
        :param List[str] commands: The command to be run in the image.
        :param List['ServiceTaskSpecContainerSpecConfigArgs'] configs: See Configs below for details.
        :param str dir: The working directory for commands to run in.
        :param 'ServiceTaskSpecContainerSpecDnsConfigArgs' dns_config: See DNS Config below for details.
        :param Mapping[str, str] env: A list of environment variables in the form VAR=value.
        :param List[str] groups: A list of additional groups that the container process will run as.
        :param 'ServiceTaskSpecContainerSpecHealthcheckArgs' healthcheck: See Healthcheck below for details.
        :param str hostname: The hostname to use for the container, as a valid RFC 1123 hostname.
        :param str isolation: Isolation technology of the containers running the service. (Windows only). Valid values are: `default|process|hyperv`
        :param List['ServiceTaskSpecContainerSpecLabelArgs'] labels: See Labels below for details.
        :param List['ServiceTaskSpecContainerSpecMountArgs'] mounts: See Mounts below for details.
        :param 'ServiceTaskSpecContainerSpecPrivilegesArgs' privileges: See Privileges below for details.
        :param bool read_only: Mount the container's root filesystem as read only.
        :param List['ServiceTaskSpecContainerSpecSecretArgs'] secrets: See Secrets below for details.
        :param str stop_grace_period: Amount of time to wait for the container to terminate before forcefully removing it `(ms|s|m|h)`.
        :param str stop_signal: Signal to stop the container.
        :param str user: The user inside the container.
        """
        pulumi.set(__self__, "image", image)
        if args is not None:
            pulumi.set(__self__, "args", args)
        if commands is not None:
            pulumi.set(__self__, "commands", commands)
        if configs is not None:
            pulumi.set(__self__, "configs", configs)
        if dir is not None:
            pulumi.set(__self__, "dir", dir)
        if dns_config is not None:
            pulumi.set(__self__, "dns_config", dns_config)
        if env is not None:
            pulumi.set(__self__, "env", env)
        if groups is not None:
            pulumi.set(__self__, "groups", groups)
        if healthcheck is not None:
            pulumi.set(__self__, "healthcheck", healthcheck)
        if hostname is not None:
            pulumi.set(__self__, "hostname", hostname)
        if hosts is not None:
            pulumi.set(__self__, "hosts", hosts)
        if isolation is not None:
            pulumi.set(__self__, "isolation", isolation)
        if labels is not None:
            pulumi.set(__self__, "labels", labels)
        if mounts is not None:
            pulumi.set(__self__, "mounts", mounts)
        if privileges is not None:
            pulumi.set(__self__, "privileges", privileges)
        if read_only is not None:
            pulumi.set(__self__, "read_only", read_only)
        if secrets is not None:
            pulumi.set(__self__, "secrets", secrets)
        if stop_grace_period is not None:
            pulumi.set(__self__, "stop_grace_period", stop_grace_period)
        if stop_signal is not None:
            pulumi.set(__self__, "stop_signal", stop_signal)
        if user is not None:
            pulumi.set(__self__, "user", user)

    @property
    @pulumi.getter
    def image(self) -> str:
        """
        The image used to create the Docker service.
        """
        return pulumi.get(self, "image")

    @property
    @pulumi.getter
    def args(self) -> Optional[List[str]]:
        """
        Arguments to the command.
        """
        return pulumi.get(self, "args")

    @property
    @pulumi.getter
    def commands(self) -> Optional[List[str]]:
        """
        The command to be run in the image.
        """
        return pulumi.get(self, "commands")

    @property
    @pulumi.getter
    def configs(self) -> Optional[List['outputs.ServiceTaskSpecContainerSpecConfig']]:
        """
        See Configs below for details.
        """
        return pulumi.get(self, "configs")

    @property
    @pulumi.getter
    def dir(self) -> Optional[str]:
        """
        The working directory for commands to run in.
        """
        return pulumi.get(self, "dir")

    @property
    @pulumi.getter(name="dnsConfig")
    def dns_config(self) -> Optional['outputs.ServiceTaskSpecContainerSpecDnsConfig']:
        """
        See DNS Config below for details.
        """
        return pulumi.get(self, "dns_config")

    @property
    @pulumi.getter
    def env(self) -> Optional[Mapping[str, str]]:
        """
        A list of environment variables in the form VAR=value.
        """
        return pulumi.get(self, "env")

    @property
    @pulumi.getter
    def groups(self) -> Optional[List[str]]:
        """
        A list of additional groups that the container process will run as.
        """
        return pulumi.get(self, "groups")

    @property
    @pulumi.getter
    def healthcheck(self) -> Optional['outputs.ServiceTaskSpecContainerSpecHealthcheck']:
        """
        See Healthcheck below for details.
        """
        return pulumi.get(self, "healthcheck")

    @property
    @pulumi.getter
    def hostname(self) -> Optional[str]:
        """
        The hostname to use for the container, as a valid RFC 1123 hostname.
        """
        return pulumi.get(self, "hostname")

    @property
    @pulumi.getter
    def hosts(self) -> Optional[List['outputs.ServiceTaskSpecContainerSpecHost']]:
        return pulumi.get(self, "hosts")

    @property
    @pulumi.getter
    def isolation(self) -> Optional[str]:
        """
        Isolation technology of the containers running the service. (Windows only). Valid values are: `default|process|hyperv`
        """
        return pulumi.get(self, "isolation")

    @property
    @pulumi.getter
    def labels(self) -> Optional[List['outputs.ServiceTaskSpecContainerSpecLabel']]:
        """
        See Labels below for details.
        """
        return pulumi.get(self, "labels")

    @property
    @pulumi.getter
    def mounts(self) -> Optional[List['outputs.ServiceTaskSpecContainerSpecMount']]:
        """
        See Mounts below for details.
        """
        return pulumi.get(self, "mounts")

    @property
    @pulumi.getter
    def privileges(self) -> Optional['outputs.ServiceTaskSpecContainerSpecPrivileges']:
        """
        See Privileges below for details.
        """
        return pulumi.get(self, "privileges")

    @property
    @pulumi.getter(name="readOnly")
    def read_only(self) -> Optional[bool]:
        """
        Mount the container's root filesystem as read only.
        """
        return pulumi.get(self, "read_only")

    @property
    @pulumi.getter
    def secrets(self) -> Optional[List['outputs.ServiceTaskSpecContainerSpecSecret']]:
        """
        See Secrets below for details.
        """
        return pulumi.get(self, "secrets")

    @property
    @pulumi.getter(name="stopGracePeriod")
    def stop_grace_period(self) -> Optional[str]:
        """
        Amount of time to wait for the container to terminate before forcefully removing it `(ms|s|m|h)`.
        """
        return pulumi.get(self, "stop_grace_period")

    @property
    @pulumi.getter(name="stopSignal")
    def stop_signal(self) -> Optional[str]:
        """
        Signal to stop the container.
        """
        return pulumi.get(self, "stop_signal")

    @property
    @pulumi.getter
    def user(self) -> Optional[str]:
        """
        The user inside the container.
        """
        return pulumi.get(self, "user")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecConfig(dict):
    def __init__(__self__, *,
                 config_id: str,
                 file_name: str,
                 config_name: Optional[str] = None,
                 file_gid: Optional[str] = None,
                 file_mode: Optional[float] = None,
                 file_uid: Optional[str] = None):
        """
        :param str config_id: ConfigID represents the ID of the specific config.
        :param str file_name: Represents the final filename in the filesystem. The specific target file that the config data is written within the docker container, e.g. `/root/config/config.json`
        :param str config_name: The name of the config that this references, but internally it is just provided for lookup/display purposes
        :param str file_gid: Represents the file GID. Defaults: `0`
        :param float file_mode: Represents the FileMode of the file. Defaults: `0444`
        :param str file_uid: Represents the file UID. Defaults: `0`
        """
        pulumi.set(__self__, "config_id", config_id)
        pulumi.set(__self__, "file_name", file_name)
        if config_name is not None:
            pulumi.set(__self__, "config_name", config_name)
        if file_gid is not None:
            pulumi.set(__self__, "file_gid", file_gid)
        if file_mode is not None:
            pulumi.set(__self__, "file_mode", file_mode)
        if file_uid is not None:
            pulumi.set(__self__, "file_uid", file_uid)

    @property
    @pulumi.getter(name="configId")
    def config_id(self) -> str:
        """
        ConfigID represents the ID of the specific config.
        """
        return pulumi.get(self, "config_id")

    @property
    @pulumi.getter(name="fileName")
    def file_name(self) -> str:
        """
        Represents the final filename in the filesystem. The specific target file that the config data is written within the docker container, e.g. `/root/config/config.json`
        """
        return pulumi.get(self, "file_name")

    @property
    @pulumi.getter(name="configName")
    def config_name(self) -> Optional[str]:
        """
        The name of the config that this references, but internally it is just provided for lookup/display purposes
        """
        return pulumi.get(self, "config_name")

    @property
    @pulumi.getter(name="fileGid")
    def file_gid(self) -> Optional[str]:
        """
        Represents the file GID. Defaults: `0`
        """
        return pulumi.get(self, "file_gid")

    @property
    @pulumi.getter(name="fileMode")
    def file_mode(self) -> Optional[float]:
        """
        Represents the FileMode of the file. Defaults: `0444`
        """
        return pulumi.get(self, "file_mode")

    @property
    @pulumi.getter(name="fileUid")
    def file_uid(self) -> Optional[str]:
        """
        Represents the file UID. Defaults: `0`
        """
        return pulumi.get(self, "file_uid")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecDnsConfig(dict):
    def __init__(__self__, *,
                 nameservers: List[str],
                 options: Optional[List[str]] = None,
                 searches: Optional[List[str]] = None):
        """
        :param List[str] nameservers: The IP addresses of the name servers, for example, `8.8.8.8`
        :param List[str] options: A list of internal resolver variables to be modified, for example, `debug`, `ndots:3`
        :param List[str] searches: A search list for host-name lookup.
        """
        pulumi.set(__self__, "nameservers", nameservers)
        if options is not None:
            pulumi.set(__self__, "options", options)
        if searches is not None:
            pulumi.set(__self__, "searches", searches)

    @property
    @pulumi.getter
    def nameservers(self) -> List[str]:
        """
        The IP addresses of the name servers, for example, `8.8.8.8`
        """
        return pulumi.get(self, "nameservers")

    @property
    @pulumi.getter
    def options(self) -> Optional[List[str]]:
        """
        A list of internal resolver variables to be modified, for example, `debug`, `ndots:3`
        """
        return pulumi.get(self, "options")

    @property
    @pulumi.getter
    def searches(self) -> Optional[List[str]]:
        """
        A search list for host-name lookup.
        """
        return pulumi.get(self, "searches")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecHealthcheck(dict):
    def __init__(__self__, *,
                 tests: List[str],
                 interval: Optional[str] = None,
                 retries: Optional[float] = None,
                 start_period: Optional[str] = None,
                 timeout: Optional[str] = None):
        """
        :param List[str] tests: Command to run to check health. For example, to run `curl -f http://localhost/health` set the
               command to be `["CMD", "curl", "-f", "http://localhost/health"]`.
        :param str interval: Time between running the check `(ms|s|m|h)`. Default: `0s`.
        :param float retries: Consecutive failures needed to report unhealthy. Default: `0`.
        :param str start_period: Start period for the container to initialize before counting retries towards unstable `(ms|s|m|h)`. Default: `0s`.
        :param str timeout: Maximum time to allow one check to run `(ms|s|m|h)`. Default: `0s`.
        """
        pulumi.set(__self__, "tests", tests)
        if interval is not None:
            pulumi.set(__self__, "interval", interval)
        if retries is not None:
            pulumi.set(__self__, "retries", retries)
        if start_period is not None:
            pulumi.set(__self__, "start_period", start_period)
        if timeout is not None:
            pulumi.set(__self__, "timeout", timeout)

    @property
    @pulumi.getter
    def tests(self) -> List[str]:
        """
        Command to run to check health. For example, to run `curl -f http://localhost/health` set the
        command to be `["CMD", "curl", "-f", "http://localhost/health"]`.
        """
        return pulumi.get(self, "tests")

    @property
    @pulumi.getter
    def interval(self) -> Optional[str]:
        """
        Time between running the check `(ms|s|m|h)`. Default: `0s`.
        """
        return pulumi.get(self, "interval")

    @property
    @pulumi.getter
    def retries(self) -> Optional[float]:
        """
        Consecutive failures needed to report unhealthy. Default: `0`.
        """
        return pulumi.get(self, "retries")

    @property
    @pulumi.getter(name="startPeriod")
    def start_period(self) -> Optional[str]:
        """
        Start period for the container to initialize before counting retries towards unstable `(ms|s|m|h)`. Default: `0s`.
        """
        return pulumi.get(self, "start_period")

    @property
    @pulumi.getter
    def timeout(self) -> Optional[str]:
        """
        Maximum time to allow one check to run `(ms|s|m|h)`. Default: `0s`.
        """
        return pulumi.get(self, "timeout")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecHost(dict):
    def __init__(__self__, *,
                 host: str,
                 ip: str):
        """
        :param str host: A list of hostname/IP mappings to add to the container's hosts file.
        :param str ip: The ip
        """
        pulumi.set(__self__, "host", host)
        pulumi.set(__self__, "ip", ip)

    @property
    @pulumi.getter
    def host(self) -> str:
        """
        A list of hostname/IP mappings to add to the container's hosts file.
        """
        return pulumi.get(self, "host")

    @property
    @pulumi.getter
    def ip(self) -> str:
        """
        The ip
        """
        return pulumi.get(self, "ip")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        """
        :param str label: Name of the label
        :param str value: Value of the label
        """
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        """
        Name of the label
        """
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        """
        Value of the label
        """
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecMount(dict):
    def __init__(__self__, *,
                 target: str,
                 type: str,
                 bind_options: Optional['outputs.ServiceTaskSpecContainerSpecMountBindOptions'] = None,
                 read_only: Optional[bool] = None,
                 source: Optional[str] = None,
                 tmpfs_options: Optional['outputs.ServiceTaskSpecContainerSpecMountTmpfsOptions'] = None,
                 volume_options: Optional['outputs.ServiceTaskSpecContainerSpecMountVolumeOptions'] = None):
        """
        :param str target: The container path.
        :param str type: SELinux type label
        :param 'ServiceTaskSpecContainerSpecMountBindOptionsArgs' bind_options: Optional configuration for the `bind` type.
        :param bool read_only: Mount the container's root filesystem as read only.
        :param str source: The mount source (e.g., a volume name, a host path)
        :param 'ServiceTaskSpecContainerSpecMountTmpfsOptionsArgs' tmpfs_options: Optional configuration for the `tmpf` type.
        :param 'ServiceTaskSpecContainerSpecMountVolumeOptionsArgs' volume_options: Optional configuration for the `volume` type.
        """
        pulumi.set(__self__, "target", target)
        pulumi.set(__self__, "type", type)
        if bind_options is not None:
            pulumi.set(__self__, "bind_options", bind_options)
        if read_only is not None:
            pulumi.set(__self__, "read_only", read_only)
        if source is not None:
            pulumi.set(__self__, "source", source)
        if tmpfs_options is not None:
            pulumi.set(__self__, "tmpfs_options", tmpfs_options)
        if volume_options is not None:
            pulumi.set(__self__, "volume_options", volume_options)

    @property
    @pulumi.getter
    def target(self) -> str:
        """
        The container path.
        """
        return pulumi.get(self, "target")

    @property
    @pulumi.getter
    def type(self) -> str:
        """
        SELinux type label
        """
        return pulumi.get(self, "type")

    @property
    @pulumi.getter(name="bindOptions")
    def bind_options(self) -> Optional['outputs.ServiceTaskSpecContainerSpecMountBindOptions']:
        """
        Optional configuration for the `bind` type.
        """
        return pulumi.get(self, "bind_options")

    @property
    @pulumi.getter(name="readOnly")
    def read_only(self) -> Optional[bool]:
        """
        Mount the container's root filesystem as read only.
        """
        return pulumi.get(self, "read_only")

    @property
    @pulumi.getter
    def source(self) -> Optional[str]:
        """
        The mount source (e.g., a volume name, a host path)
        """
        return pulumi.get(self, "source")

    @property
    @pulumi.getter(name="tmpfsOptions")
    def tmpfs_options(self) -> Optional['outputs.ServiceTaskSpecContainerSpecMountTmpfsOptions']:
        """
        Optional configuration for the `tmpf` type.
        """
        return pulumi.get(self, "tmpfs_options")

    @property
    @pulumi.getter(name="volumeOptions")
    def volume_options(self) -> Optional['outputs.ServiceTaskSpecContainerSpecMountVolumeOptions']:
        """
        Optional configuration for the `volume` type.
        """
        return pulumi.get(self, "volume_options")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecMountBindOptions(dict):
    def __init__(__self__, *,
                 propagation: Optional[str] = None):
        """
        :param str propagation: A propagation mode with the value.
        """
        if propagation is not None:
            pulumi.set(__self__, "propagation", propagation)

    @property
    @pulumi.getter
    def propagation(self) -> Optional[str]:
        """
        A propagation mode with the value.
        """
        return pulumi.get(self, "propagation")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecMountTmpfsOptions(dict):
    def __init__(__self__, *,
                 mode: Optional[float] = None,
                 size_bytes: Optional[float] = None):
        """
        :param float mode: See Mode below for details.
        :param float size_bytes: The size for the tmpfs mount in bytes.
        """
        if mode is not None:
            pulumi.set(__self__, "mode", mode)
        if size_bytes is not None:
            pulumi.set(__self__, "size_bytes", size_bytes)

    @property
    @pulumi.getter
    def mode(self) -> Optional[float]:
        """
        See Mode below for details.
        """
        return pulumi.get(self, "mode")

    @property
    @pulumi.getter(name="sizeBytes")
    def size_bytes(self) -> Optional[float]:
        """
        The size for the tmpfs mount in bytes.
        """
        return pulumi.get(self, "size_bytes")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecMountVolumeOptions(dict):
    def __init__(__self__, *,
                 driver_name: Optional[str] = None,
                 driver_options: Optional[Mapping[str, str]] = None,
                 labels: Optional[List['outputs.ServiceTaskSpecContainerSpecMountVolumeOptionsLabel']] = None,
                 no_copy: Optional[bool] = None):
        """
        :param List['ServiceTaskSpecContainerSpecMountVolumeOptionsLabelArgs'] labels: See Labels below for details.
        :param bool no_copy: Whether to populate volume with data from the target.
        """
        if driver_name is not None:
            pulumi.set(__self__, "driver_name", driver_name)
        if driver_options is not None:
            pulumi.set(__self__, "driver_options", driver_options)
        if labels is not None:
            pulumi.set(__self__, "labels", labels)
        if no_copy is not None:
            pulumi.set(__self__, "no_copy", no_copy)

    @property
    @pulumi.getter(name="driverName")
    def driver_name(self) -> Optional[str]:
        return pulumi.get(self, "driver_name")

    @property
    @pulumi.getter(name="driverOptions")
    def driver_options(self) -> Optional[Mapping[str, str]]:
        return pulumi.get(self, "driver_options")

    @property
    @pulumi.getter
    def labels(self) -> Optional[List['outputs.ServiceTaskSpecContainerSpecMountVolumeOptionsLabel']]:
        """
        See Labels below for details.
        """
        return pulumi.get(self, "labels")

    @property
    @pulumi.getter(name="noCopy")
    def no_copy(self) -> Optional[bool]:
        """
        Whether to populate volume with data from the target.
        """
        return pulumi.get(self, "no_copy")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecMountVolumeOptionsLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        """
        :param str label: Name of the label
        :param str value: Value of the label
        """
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        """
        Name of the label
        """
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        """
        Value of the label
        """
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecPrivileges(dict):
    def __init__(__self__, *,
                 credential_spec: Optional['outputs.ServiceTaskSpecContainerSpecPrivilegesCredentialSpec'] = None,
                 se_linux_context: Optional['outputs.ServiceTaskSpecContainerSpecPrivilegesSeLinuxContext'] = None):
        """
        :param 'ServiceTaskSpecContainerSpecPrivilegesCredentialSpecArgs' credential_spec: For managed service account (Windows only)
        :param 'ServiceTaskSpecContainerSpecPrivilegesSeLinuxContextArgs' se_linux_context: SELinux labels of the container
        """
        if credential_spec is not None:
            pulumi.set(__self__, "credential_spec", credential_spec)
        if se_linux_context is not None:
            pulumi.set(__self__, "se_linux_context", se_linux_context)

    @property
    @pulumi.getter(name="credentialSpec")
    def credential_spec(self) -> Optional['outputs.ServiceTaskSpecContainerSpecPrivilegesCredentialSpec']:
        """
        For managed service account (Windows only)
        """
        return pulumi.get(self, "credential_spec")

    @property
    @pulumi.getter(name="seLinuxContext")
    def se_linux_context(self) -> Optional['outputs.ServiceTaskSpecContainerSpecPrivilegesSeLinuxContext']:
        """
        SELinux labels of the container
        """
        return pulumi.get(self, "se_linux_context")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecPrivilegesCredentialSpec(dict):
    def __init__(__self__, *,
                 file: Optional[str] = None,
                 registry: Optional[str] = None):
        """
        :param str file: Load credential spec from this file.
        :param str registry: Load credential spec from this value in the Windows registry.
        """
        if file is not None:
            pulumi.set(__self__, "file", file)
        if registry is not None:
            pulumi.set(__self__, "registry", registry)

    @property
    @pulumi.getter
    def file(self) -> Optional[str]:
        """
        Load credential spec from this file.
        """
        return pulumi.get(self, "file")

    @property
    @pulumi.getter
    def registry(self) -> Optional[str]:
        """
        Load credential spec from this value in the Windows registry.
        """
        return pulumi.get(self, "registry")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecPrivilegesSeLinuxContext(dict):
    def __init__(__self__, *,
                 disable: Optional[bool] = None,
                 level: Optional[str] = None,
                 role: Optional[str] = None,
                 type: Optional[str] = None,
                 user: Optional[str] = None):
        """
        :param bool disable: Disable SELinux
        :param str level: SELinux level label
        :param str role: SELinux role label
        :param str type: SELinux type label
        :param str user: The user inside the container.
        """
        if disable is not None:
            pulumi.set(__self__, "disable", disable)
        if level is not None:
            pulumi.set(__self__, "level", level)
        if role is not None:
            pulumi.set(__self__, "role", role)
        if type is not None:
            pulumi.set(__self__, "type", type)
        if user is not None:
            pulumi.set(__self__, "user", user)

    @property
    @pulumi.getter
    def disable(self) -> Optional[bool]:
        """
        Disable SELinux
        """
        return pulumi.get(self, "disable")

    @property
    @pulumi.getter
    def level(self) -> Optional[str]:
        """
        SELinux level label
        """
        return pulumi.get(self, "level")

    @property
    @pulumi.getter
    def role(self) -> Optional[str]:
        """
        SELinux role label
        """
        return pulumi.get(self, "role")

    @property
    @pulumi.getter
    def type(self) -> Optional[str]:
        """
        SELinux type label
        """
        return pulumi.get(self, "type")

    @property
    @pulumi.getter
    def user(self) -> Optional[str]:
        """
        The user inside the container.
        """
        return pulumi.get(self, "user")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecContainerSpecSecret(dict):
    def __init__(__self__, *,
                 file_name: str,
                 secret_id: str,
                 file_gid: Optional[str] = None,
                 file_mode: Optional[float] = None,
                 file_uid: Optional[str] = None,
                 secret_name: Optional[str] = None):
        """
        :param str file_name: Represents the final filename in the filesystem. The specific target file that the secret data is written within the docker container, e.g. `/root/secret/secret.json`
        :param str secret_id: ConfigID represents the ID of the specific secret.
        :param str file_gid: Represents the file GID. Defaults: `0`
        :param float file_mode: Represents the FileMode of the file. Defaults: `0444`
        :param str file_uid: Represents the file UID. Defaults: `0`
        :param str secret_name: The name of the secret that this references, but internally it is just provided for lookup/display purposes
        """
        pulumi.set(__self__, "file_name", file_name)
        pulumi.set(__self__, "secret_id", secret_id)
        if file_gid is not None:
            pulumi.set(__self__, "file_gid", file_gid)
        if file_mode is not None:
            pulumi.set(__self__, "file_mode", file_mode)
        if file_uid is not None:
            pulumi.set(__self__, "file_uid", file_uid)
        if secret_name is not None:
            pulumi.set(__self__, "secret_name", secret_name)

    @property
    @pulumi.getter(name="fileName")
    def file_name(self) -> str:
        """
        Represents the final filename in the filesystem. The specific target file that the secret data is written within the docker container, e.g. `/root/secret/secret.json`
        """
        return pulumi.get(self, "file_name")

    @property
    @pulumi.getter(name="secretId")
    def secret_id(self) -> str:
        """
        ConfigID represents the ID of the specific secret.
        """
        return pulumi.get(self, "secret_id")

    @property
    @pulumi.getter(name="fileGid")
    def file_gid(self) -> Optional[str]:
        """
        Represents the file GID. Defaults: `0`
        """
        return pulumi.get(self, "file_gid")

    @property
    @pulumi.getter(name="fileMode")
    def file_mode(self) -> Optional[float]:
        """
        Represents the FileMode of the file. Defaults: `0444`
        """
        return pulumi.get(self, "file_mode")

    @property
    @pulumi.getter(name="fileUid")
    def file_uid(self) -> Optional[str]:
        """
        Represents the file UID. Defaults: `0`
        """
        return pulumi.get(self, "file_uid")

    @property
    @pulumi.getter(name="secretName")
    def secret_name(self) -> Optional[str]:
        """
        The name of the secret that this references, but internally it is just provided for lookup/display purposes
        """
        return pulumi.get(self, "secret_name")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecLogDriver(dict):
    def __init__(__self__, *,
                 name: str,
                 options: Optional[Mapping[str, str]] = None):
        """
        :param str name: The logging driver to use. Either `(none|json-file|syslog|journald|gelf|fluentd|awslogs|splunk|etwlogs|gcplogs)`.
        :param Mapping[str, str] options: The options for the logging driver, e.g.
        """
        pulumi.set(__self__, "name", name)
        if options is not None:
            pulumi.set(__self__, "options", options)

    @property
    @pulumi.getter
    def name(self) -> str:
        """
        The logging driver to use. Either `(none|json-file|syslog|journald|gelf|fluentd|awslogs|splunk|etwlogs|gcplogs)`.
        """
        return pulumi.get(self, "name")

    @property
    @pulumi.getter
    def options(self) -> Optional[Mapping[str, str]]:
        """
        The options for the logging driver, e.g.
        """
        return pulumi.get(self, "options")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecPlacement(dict):
    def __init__(__self__, *,
                 constraints: Optional[List[str]] = None,
                 platforms: Optional[List['outputs.ServiceTaskSpecPlacementPlatform']] = None,
                 prefs: Optional[List[str]] = None):
        """
        :param List[str] constraints: An array of constraints. e.g.: `node.role==manager`
        :param List['ServiceTaskSpecPlacementPlatformArgs'] platforms: Platforms stores all the platforms that the service's image can run on
        :param List[str] prefs: Preferences provide a way to make the scheduler aware of factors such as topology. They are provided in order from highest to lowest precedence, e.g.: `spread=node.role.manager`
        """
        if constraints is not None:
            pulumi.set(__self__, "constraints", constraints)
        if platforms is not None:
            pulumi.set(__self__, "platforms", platforms)
        if prefs is not None:
            pulumi.set(__self__, "prefs", prefs)

    @property
    @pulumi.getter
    def constraints(self) -> Optional[List[str]]:
        """
        An array of constraints. e.g.: `node.role==manager`
        """
        return pulumi.get(self, "constraints")

    @property
    @pulumi.getter
    def platforms(self) -> Optional[List['outputs.ServiceTaskSpecPlacementPlatform']]:
        """
        Platforms stores all the platforms that the service's image can run on
        """
        return pulumi.get(self, "platforms")

    @property
    @pulumi.getter
    def prefs(self) -> Optional[List[str]]:
        """
        Preferences provide a way to make the scheduler aware of factors such as topology. They are provided in order from highest to lowest precedence, e.g.: `spread=node.role.manager`
        """
        return pulumi.get(self, "prefs")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecPlacementPlatform(dict):
    def __init__(__self__, *,
                 architecture: str,
                 os: str):
        """
        :param str architecture: The architecture, e.g., `amd64`
        :param str os: The operation system, e.g., `linux`
        """
        pulumi.set(__self__, "architecture", architecture)
        pulumi.set(__self__, "os", os)

    @property
    @pulumi.getter
    def architecture(self) -> str:
        """
        The architecture, e.g., `amd64`
        """
        return pulumi.get(self, "architecture")

    @property
    @pulumi.getter
    def os(self) -> str:
        """
        The operation system, e.g., `linux`
        """
        return pulumi.get(self, "os")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecResources(dict):
    def __init__(__self__, *,
                 limits: Optional['outputs.ServiceTaskSpecResourcesLimits'] = None,
                 reservation: Optional['outputs.ServiceTaskSpecResourcesReservation'] = None):
        """
        :param 'ServiceTaskSpecResourcesLimitsArgs' limits: Describes the resources which can be advertised by a node and requested by a task.
        :param 'ServiceTaskSpecResourcesReservationArgs' reservation: An object describing the resources which can be advertised by a node and requested by a task.
        """
        if limits is not None:
            pulumi.set(__self__, "limits", limits)
        if reservation is not None:
            pulumi.set(__self__, "reservation", reservation)

    @property
    @pulumi.getter
    def limits(self) -> Optional['outputs.ServiceTaskSpecResourcesLimits']:
        """
        Describes the resources which can be advertised by a node and requested by a task.
        """
        return pulumi.get(self, "limits")

    @property
    @pulumi.getter
    def reservation(self) -> Optional['outputs.ServiceTaskSpecResourcesReservation']:
        """
        An object describing the resources which can be advertised by a node and requested by a task.
        """
        return pulumi.get(self, "reservation")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecResourcesLimits(dict):
    def __init__(__self__, *,
                 generic_resources: Optional['outputs.ServiceTaskSpecResourcesLimitsGenericResources'] = None,
                 memory_bytes: Optional[float] = None,
                 nano_cpus: Optional[float] = None):
        """
        :param 'ServiceTaskSpecResourcesLimitsGenericResourcesArgs' generic_resources: User-defined resources can be either Integer resources (e.g, SSD=3) or String resources (e.g, GPU=UUID1)
        :param float memory_bytes: The amount of memory in bytes the container allocates
        :param float nano_cpus: CPU shares in units of 1/1e9 (or 10^-9) of the CPU. Should be at least 1000000
        """
        if generic_resources is not None:
            pulumi.set(__self__, "generic_resources", generic_resources)
        if memory_bytes is not None:
            pulumi.set(__self__, "memory_bytes", memory_bytes)
        if nano_cpus is not None:
            pulumi.set(__self__, "nano_cpus", nano_cpus)

    @property
    @pulumi.getter(name="genericResources")
    def generic_resources(self) -> Optional['outputs.ServiceTaskSpecResourcesLimitsGenericResources']:
        """
        User-defined resources can be either Integer resources (e.g, SSD=3) or String resources (e.g, GPU=UUID1)
        """
        return pulumi.get(self, "generic_resources")

    @property
    @pulumi.getter(name="memoryBytes")
    def memory_bytes(self) -> Optional[float]:
        """
        The amount of memory in bytes the container allocates
        """
        return pulumi.get(self, "memory_bytes")

    @property
    @pulumi.getter(name="nanoCpus")
    def nano_cpus(self) -> Optional[float]:
        """
        CPU shares in units of 1/1e9 (or 10^-9) of the CPU. Should be at least 1000000
        """
        return pulumi.get(self, "nano_cpus")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecResourcesLimitsGenericResources(dict):
    def __init__(__self__, *,
                 discrete_resources_specs: Optional[List[str]] = None,
                 named_resources_specs: Optional[List[str]] = None):
        """
        :param List[str] discrete_resources_specs: The Integer resources, delimited by `=`
        :param List[str] named_resources_specs: The String resources, delimited by `=`
        """
        if discrete_resources_specs is not None:
            pulumi.set(__self__, "discrete_resources_specs", discrete_resources_specs)
        if named_resources_specs is not None:
            pulumi.set(__self__, "named_resources_specs", named_resources_specs)

    @property
    @pulumi.getter(name="discreteResourcesSpecs")
    def discrete_resources_specs(self) -> Optional[List[str]]:
        """
        The Integer resources, delimited by `=`
        """
        return pulumi.get(self, "discrete_resources_specs")

    @property
    @pulumi.getter(name="namedResourcesSpecs")
    def named_resources_specs(self) -> Optional[List[str]]:
        """
        The String resources, delimited by `=`
        """
        return pulumi.get(self, "named_resources_specs")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecResourcesReservation(dict):
    def __init__(__self__, *,
                 generic_resources: Optional['outputs.ServiceTaskSpecResourcesReservationGenericResources'] = None,
                 memory_bytes: Optional[float] = None,
                 nano_cpus: Optional[float] = None):
        """
        :param 'ServiceTaskSpecResourcesReservationGenericResourcesArgs' generic_resources: User-defined resources can be either Integer resources (e.g, SSD=3) or String resources (e.g, GPU=UUID1)
        :param float memory_bytes: The amount of memory in bytes the container allocates
        :param float nano_cpus: CPU shares in units of 1/1e9 (or 10^-9) of the CPU. Should be at least 1000000
        """
        if generic_resources is not None:
            pulumi.set(__self__, "generic_resources", generic_resources)
        if memory_bytes is not None:
            pulumi.set(__self__, "memory_bytes", memory_bytes)
        if nano_cpus is not None:
            pulumi.set(__self__, "nano_cpus", nano_cpus)

    @property
    @pulumi.getter(name="genericResources")
    def generic_resources(self) -> Optional['outputs.ServiceTaskSpecResourcesReservationGenericResources']:
        """
        User-defined resources can be either Integer resources (e.g, SSD=3) or String resources (e.g, GPU=UUID1)
        """
        return pulumi.get(self, "generic_resources")

    @property
    @pulumi.getter(name="memoryBytes")
    def memory_bytes(self) -> Optional[float]:
        """
        The amount of memory in bytes the container allocates
        """
        return pulumi.get(self, "memory_bytes")

    @property
    @pulumi.getter(name="nanoCpus")
    def nano_cpus(self) -> Optional[float]:
        """
        CPU shares in units of 1/1e9 (or 10^-9) of the CPU. Should be at least 1000000
        """
        return pulumi.get(self, "nano_cpus")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecResourcesReservationGenericResources(dict):
    def __init__(__self__, *,
                 discrete_resources_specs: Optional[List[str]] = None,
                 named_resources_specs: Optional[List[str]] = None):
        """
        :param List[str] discrete_resources_specs: The Integer resources, delimited by `=`
        :param List[str] named_resources_specs: The String resources, delimited by `=`
        """
        if discrete_resources_specs is not None:
            pulumi.set(__self__, "discrete_resources_specs", discrete_resources_specs)
        if named_resources_specs is not None:
            pulumi.set(__self__, "named_resources_specs", named_resources_specs)

    @property
    @pulumi.getter(name="discreteResourcesSpecs")
    def discrete_resources_specs(self) -> Optional[List[str]]:
        """
        The Integer resources, delimited by `=`
        """
        return pulumi.get(self, "discrete_resources_specs")

    @property
    @pulumi.getter(name="namedResourcesSpecs")
    def named_resources_specs(self) -> Optional[List[str]]:
        """
        The String resources, delimited by `=`
        """
        return pulumi.get(self, "named_resources_specs")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceTaskSpecRestartPolicy(dict):
    def __init__(__self__, *,
                 condition: Optional[str] = None,
                 delay: Optional[str] = None,
                 max_attempts: Optional[float] = None,
                 window: Optional[str] = None):
        """
        :param str condition: Condition for restart: `(none|on-failure|any)`
        :param str delay: Delay between restart attempts `(ms|s|m|h)`
        :param float max_attempts: Maximum attempts to restart a given container before giving up (default value is `0`, which is ignored)
        :param str window: The time window used to evaluate the restart policy (default value is `0`, which is unbounded) `(ms|s|m|h)`
        """
        if condition is not None:
            pulumi.set(__self__, "condition", condition)
        if delay is not None:
            pulumi.set(__self__, "delay", delay)
        if max_attempts is not None:
            pulumi.set(__self__, "max_attempts", max_attempts)
        if window is not None:
            pulumi.set(__self__, "window", window)

    @property
    @pulumi.getter
    def condition(self) -> Optional[str]:
        """
        Condition for restart: `(none|on-failure|any)`
        """
        return pulumi.get(self, "condition")

    @property
    @pulumi.getter
    def delay(self) -> Optional[str]:
        """
        Delay between restart attempts `(ms|s|m|h)`
        """
        return pulumi.get(self, "delay")

    @property
    @pulumi.getter(name="maxAttempts")
    def max_attempts(self) -> Optional[float]:
        """
        Maximum attempts to restart a given container before giving up (default value is `0`, which is ignored)
        """
        return pulumi.get(self, "max_attempts")

    @property
    @pulumi.getter
    def window(self) -> Optional[str]:
        """
        The time window used to evaluate the restart policy (default value is `0`, which is unbounded) `(ms|s|m|h)`
        """
        return pulumi.get(self, "window")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class ServiceUpdateConfig(dict):
    def __init__(__self__, *,
                 delay: Optional[str] = None,
                 failure_action: Optional[str] = None,
                 max_failure_ratio: Optional[str] = None,
                 monitor: Optional[str] = None,
                 order: Optional[str] = None,
                 parallelism: Optional[float] = None):
        """
        :param str delay: Delay between updates `(ns|us|ms|s|m|h)`, e.g. `5s`.
        :param str failure_action: Action on update failure: `pause|continue|rollback`.
        :param str max_failure_ratio: The failure rate to tolerate during an update as `float`. **Important:** the `float`need to be wrapped in a `string` to avoid internal
               casting and precision errors.
        :param str monitor: Duration after each task update to monitor for failure `(ns|us|ms|s|m|h)`
        :param str order: Update order either 'stop-first' or 'start-first'.
        :param float parallelism: The maximum number of tasks to be updated in one iteration simultaneously (0 to update all at once).
        """
        if delay is not None:
            pulumi.set(__self__, "delay", delay)
        if failure_action is not None:
            pulumi.set(__self__, "failure_action", failure_action)
        if max_failure_ratio is not None:
            pulumi.set(__self__, "max_failure_ratio", max_failure_ratio)
        if monitor is not None:
            pulumi.set(__self__, "monitor", monitor)
        if order is not None:
            pulumi.set(__self__, "order", order)
        if parallelism is not None:
            pulumi.set(__self__, "parallelism", parallelism)

    @property
    @pulumi.getter
    def delay(self) -> Optional[str]:
        """
        Delay between updates `(ns|us|ms|s|m|h)`, e.g. `5s`.
        """
        return pulumi.get(self, "delay")

    @property
    @pulumi.getter(name="failureAction")
    def failure_action(self) -> Optional[str]:
        """
        Action on update failure: `pause|continue|rollback`.
        """
        return pulumi.get(self, "failure_action")

    @property
    @pulumi.getter(name="maxFailureRatio")
    def max_failure_ratio(self) -> Optional[str]:
        """
        The failure rate to tolerate during an update as `float`. **Important:** the `float`need to be wrapped in a `string` to avoid internal
        casting and precision errors.
        """
        return pulumi.get(self, "max_failure_ratio")

    @property
    @pulumi.getter
    def monitor(self) -> Optional[str]:
        """
        Duration after each task update to monitor for failure `(ns|us|ms|s|m|h)`
        """
        return pulumi.get(self, "monitor")

    @property
    @pulumi.getter
    def order(self) -> Optional[str]:
        """
        Update order either 'stop-first' or 'start-first'.
        """
        return pulumi.get(self, "order")

    @property
    @pulumi.getter
    def parallelism(self) -> Optional[float]:
        """
        The maximum number of tasks to be updated in one iteration simultaneously (0 to update all at once).
        """
        return pulumi.get(self, "parallelism")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class VolumeLabel(dict):
    def __init__(__self__, *,
                 label: str,
                 value: str):
        pulumi.set(__self__, "label", label)
        pulumi.set(__self__, "value", value)

    @property
    @pulumi.getter
    def label(self) -> str:
        return pulumi.get(self, "label")

    @property
    @pulumi.getter
    def value(self) -> str:
        return pulumi.get(self, "value")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop


@pulumi.output_type
class GetNetworkIpamConfigResult(dict):
    def __init__(__self__, *,
                 aux_address: Optional[Mapping[str, Any]] = None,
                 gateway: Optional[str] = None,
                 ip_range: Optional[str] = None,
                 subnet: Optional[str] = None):
        if aux_address is not None:
            pulumi.set(__self__, "aux_address", aux_address)
        if gateway is not None:
            pulumi.set(__self__, "gateway", gateway)
        if ip_range is not None:
            pulumi.set(__self__, "ip_range", ip_range)
        if subnet is not None:
            pulumi.set(__self__, "subnet", subnet)

    @property
    @pulumi.getter(name="auxAddress")
    def aux_address(self) -> Optional[Mapping[str, Any]]:
        return pulumi.get(self, "aux_address")

    @property
    @pulumi.getter
    def gateway(self) -> Optional[str]:
        return pulumi.get(self, "gateway")

    @property
    @pulumi.getter(name="ipRange")
    def ip_range(self) -> Optional[str]:
        return pulumi.get(self, "ip_range")

    @property
    @pulumi.getter
    def subnet(self) -> Optional[str]:
        return pulumi.get(self, "subnet")


@pulumi.output_type
class ProviderRegistryAuth(dict):
    def __init__(__self__, *,
                 address: str,
                 config_file: Optional[str] = None,
                 config_file_content: Optional[str] = None,
                 password: Optional[str] = None,
                 username: Optional[str] = None):
        pulumi.set(__self__, "address", address)
        if config_file is not None:
            pulumi.set(__self__, "config_file", config_file)
        if config_file_content is not None:
            pulumi.set(__self__, "config_file_content", config_file_content)
        if password is not None:
            pulumi.set(__self__, "password", password)
        if username is not None:
            pulumi.set(__self__, "username", username)

    @property
    @pulumi.getter
    def address(self) -> str:
        return pulumi.get(self, "address")

    @property
    @pulumi.getter(name="configFile")
    def config_file(self) -> Optional[str]:
        return pulumi.get(self, "config_file")

    @property
    @pulumi.getter(name="configFileContent")
    def config_file_content(self) -> Optional[str]:
        return pulumi.get(self, "config_file_content")

    @property
    @pulumi.getter
    def password(self) -> Optional[str]:
        return pulumi.get(self, "password")

    @property
    @pulumi.getter
    def username(self) -> Optional[str]:
        return pulumi.get(self, "username")

    def _translate_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop





# coding=utf-8
# *** WARNING: this file was generated by the Pulumi Terraform Bridge (tfgen) Tool. ***
# *** Do not edit by hand unless you're certain you know what you are doing! ***

import json
import warnings
import pulumi
import pulumi.runtime
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
from . import _utilities, _tables
from . import outputs

__all__ = ['Provider']


class Provider(pulumi.ProviderResource):
    def __init__(__self__,
                 resource_name: str,
                 opts: Optional[pulumi.ResourceOptions] = None,
                 ca_material: Optional[pulumi.Input[str]] = None,
                 cert_material: Optional[pulumi.Input[str]] = None,
                 cert_path: Optional[pulumi.Input[str]] = None,
                 host: Optional[pulumi.Input[str]] = None,
                 key_material: Optional[pulumi.Input[str]] = None,
                 registry_auth: Optional[pulumi.Input[List[pulumi.Input[pulumi.InputType['ProviderRegistryAuthArgs']]]]] = None,
                 __props__=None,
                 __name__=None,
                 __opts__=None):
        """
        The provider type for the docker package. By default, resources use package-wide configuration
        settings, however an explicit `Provider` instance may be created and passed during resource
        construction to achieve fine-grained programmatic control over provider settings. See the
        [documentation](https://www.pulumi.com/docs/reference/programming-model/#providers) for more information.

        :param str resource_name: The name of the resource.
        :param pulumi.ResourceOptions opts: Options for the resource.
        :param pulumi.Input[str] ca_material: PEM-encoded content of Docker host CA certificate
        :param pulumi.Input[str] cert_material: PEM-encoded content of Docker client certificate
        :param pulumi.Input[str] cert_path: Path to directory with Docker TLS config
        :param pulumi.Input[str] host: The Docker daemon address
        :param pulumi.Input[str] key_material: PEM-encoded content of Docker client private key
        """
        if __name__ is not None:
            warnings.warn("explicit use of __name__ is deprecated", DeprecationWarning)
            resource_name = __name__
        if __opts__ is not None:
            warnings.warn("explicit use of __opts__ is deprecated, use 'opts' instead", DeprecationWarning)
            opts = __opts__
        if opts is None:
            opts = pulumi.ResourceOptions()
        if not isinstance(opts, pulumi.ResourceOptions):
            raise TypeError('Expected resource options to be a ResourceOptions instance')
        if opts.version is None:
            opts.version = _utilities.get_version()
        if opts.id is None:
            if __props__ is not None:
                raise TypeError('__props__ is only valid when passed in combination with a valid opts.id to get an existing resource')
            __props__ = dict()

            __props__['ca_material'] = ca_material
            __props__['cert_material'] = cert_material
            __props__['cert_path'] = cert_path
            if host is None:
                host = (_utilities.get_env('DOCKER_HOST') or 'unix:///var/run/docker.sock')
            __props__['host'] = host
            __props__['key_material'] = key_material
            __props__['registry_auth'] = pulumi.Output.from_input(registry_auth).apply(json.dumps) if registry_auth is not None else None
        super(Provider, __self__).__init__(
            'docker',
            resource_name,
            __props__,
            opts)

    def translate_output_property(self, prop):
        return _tables.CAMEL_TO_SNAKE_CASE_TABLE.get(prop) or prop

    def translate_input_property(self, prop):
        return _tables.SNAKE_TO_CAMEL_CASE_TABLE.get(prop) or prop




# coding=utf-8
# *** WARNING: this file was generated by the Pulumi Terraform Bridge (tfgen) Tool. ***
# *** Do not edit by hand unless you're certain you know what you are doing! ***

import warnings
import pulumi
import pulumi.runtime
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
from . import _utilities, _tables

__all__ = [
    'GetRegistryImageResult',
    'AwaitableGetRegistryImageResult',
    'get_registry_image',
]

@pulumi.output_type
class GetRegistryImageResult:
    """
    A collection of values returned by getRegistryImage.
    """
    def __init__(__self__, id=None, name=None, sha256_digest=None):
        if id and not isinstance(id, str):
            raise TypeError("Expected argument 'id' to be a str")
        pulumi.set(__self__, "id", id)
        if name and not isinstance(name, str):
            raise TypeError("Expected argument 'name' to be a str")
        pulumi.set(__self__, "name", name)
        if sha256_digest and not isinstance(sha256_digest, str):
            raise TypeError("Expected argument 'sha256_digest' to be a str")
        pulumi.set(__self__, "sha256_digest", sha256_digest)

    @property
    @pulumi.getter
    def id(self) -> str:
        """
        The provider-assigned unique ID for this managed resource.
        """
        return pulumi.get(self, "id")

    @property
    @pulumi.getter
    def name(self) -> Optional[str]:
        return pulumi.get(self, "name")

    @property
    @pulumi.getter(name="sha256Digest")
    def sha256_digest(self) -> str:
        return pulumi.get(self, "sha256_digest")


class AwaitableGetRegistryImageResult(GetRegistryImageResult):
    # pylint: disable=using-constant-test
    def __await__(self):
        if False:
            yield self
        return GetRegistryImageResult(
            id=self.id,
            name=self.name,
            sha256_digest=self.sha256_digest)


def get_registry_image(name: Optional[str] = None,
                       opts: Optional[pulumi.InvokeOptions] = None) -> AwaitableGetRegistryImageResult:
    """
    Reads the image metadata from a Docker Registry. Used in conjunction with the
    [docker\_image](https://www.terraform.io/docs/providers/docker/r/image.html) resource to keep an image up
    to date on the latest available version of the tag.

    ## Example Usage

    ```python
    import pulumi
    import pulumi_docker as docker

    ubuntu_registry_image = docker.get_registry_image(name="ubuntu:precise")
    ubuntu_remote_image = docker.RemoteImage("ubuntuRemoteImage",
        name=ubuntu_registry_image.name,
        pull_triggers=[ubuntu_registry_image.sha256_digest])
    ```


    :param str name: The name of the Docker image, including any tags. e.g. `alpine:latest`
    """
    __args__ = dict()
    __args__['name'] = name
    if opts is None:
        opts = pulumi.InvokeOptions()
    if opts.version is None:
        opts.version = _utilities.get_version()
    __ret__ = pulumi.runtime.invoke('docker:index/getRegistryImage:getRegistryImage', __args__, opts=opts, typ=GetRegistryImageResult).value

    return AwaitableGetRegistryImageResult(
        id=__ret__.id,
        name=__ret__.name,
        sha256_digest=__ret__.sha256_digest)
